from __future__ import annotations

import subprocess as sp

from kraken.core import Project, Task, TaskStatus


class BufFormatTask(Task):
    description = "Format Protobuf files with buf."

    def execute(self) -> TaskStatus | None:
        command = ["buf", "format", "-w"]
        result = sp.call(command)

        return TaskStatus.from_exit_code(command, result)


class BufLintTask(Task):
    description = "Lint Protobuf files with buf."

    def execute(self) -> TaskStatus | None:
        command = ["buf", "lint"]
        result = sp.call(command, cwd=self.project.directory / "proto")

        return TaskStatus.from_exit_code(command, result)


def buf_format(*, name: str = "buf.format", project: Project | None = None) -> BufFormatTask:
    project = project or Project.current()
    return project.task(name, BufFormatTask, group="fmt")


def buf_lint(*, name: str = "buf.lint", project: Project | None = None) -> BufLintTask:
    project = project or Project.current()
    return project.task(name, BufLintTask, group="lint")



from __future__ import annotations

import logging
import os
from typing import Union, cast

from kraken.common import Supplier
from kraken.common.pyenv import get_current_venv
from kraken.core import Project, Property, Task, TaskStatus

from ..buildsystem import PythonBuildSystem
from ..settings import python_settings

logger = logging.getLogger(__name__)


class InstallTask(Task):
    build_system: Property[PythonBuildSystem | None] = Property.default(None)
    always_use_managed_env: Property[bool] = Property.default(True)
    skip_if_venv_exists: Property[bool] = Property.default(True)

    # Task

    def get_description(self) -> str | None:
        build_system = self.build_system.get()
        return (
            f"Ensure that a managed virtual environment exists. [build system: "
            f"{build_system.name if build_system else None}]"
        )

    def prepare(self) -> TaskStatus | None:
        venv = get_current_venv(os.environ)
        if not self.always_use_managed_env.get() and venv:
            return TaskStatus.skipped("using current virtual env (%s)" % venv.path)
        build_system = self.build_system.get()
        if not build_system:
            return TaskStatus.skipped("no Python build system configured")
        if not build_system.supports_managed_environments():
            return TaskStatus.skipped(
                "current build system does not supported managed environment (%s)" % type(build_system).__name__
            )
        managed_environment = build_system.get_managed_environment()
        if managed_environment.exists() and not managed_environment.always_install():
            if self.selected:
                return TaskStatus.pending("explicitly selected to run")
            if self.skip_if_venv_exists.get():
                return TaskStatus.pending(
                    "managed environment exists (%s), always install on" % managed_environment.get_path()
                )
            return TaskStatus.skipped("managed environment exists (%s)" % managed_environment.get_path())

        return TaskStatus.pending()

    def execute(self) -> TaskStatus:
        build_system = self.build_system.get()
        if not build_system:
            logger.error("no build system configured")
            return TaskStatus.failed("no build system configured")
        managed_environment = build_system.get_managed_environment()
        managed_environment.install(python_settings(self.project))
        return TaskStatus.succeeded()


def install(*, name: str = "python.install", project: Project | None = None) -> InstallTask:
    """Get or create the `python.install` task for the given project.

    The install task relies on the build system configured in the Python project settings."""

    project = project or Project.current()
    task = cast(Union[InstallTask, None], project.tasks().get(name))
    if task is None:
        task = project.task(name, InstallTask)
        task.build_system = Supplier.of_callable(lambda: python_settings(project).build_system)
        task.always_use_managed_env = Supplier.of_callable(lambda: python_settings(project).always_use_managed_env)
        task.skip_if_venv_exists = Supplier.of_callable(lambda: python_settings(project).skip_install_if_venv_exists)

    return task



from __future__ import annotations

import enum
import subprocess
from pathlib import Path
from typing import Literal

from kraken.core import Property, Task, TaskStatus
from kraken.core.system.task import TaskStatusType


def check_file_status(file_path: Path) -> CommitedFileStatus:
    file_path = file_path.absolute()
    if not file_path.exists():
        return CommitedFileStatus.MISSING
    ls_files = subprocess.run(
        ["git", "ls-files", "--", str(file_path)], capture_output=True, text=True, cwd=file_path.parent
    ).stdout.strip()
    if ls_files != file_path.name:
        return CommitedFileStatus.NOT_COMMITTED
    return CommitedFileStatus.COMMITED


class CommitedFileStatus(enum.Enum):
    """
    Represents the status of a file that is supposed to be commited.
    """

    COMMITED = enum.auto()
    NOT_COMMITTED = enum.auto()
    MISSING = enum.auto()

    def to_description(self, file_path: Path) -> str:
        match self:
            case CommitedFileStatus.COMMITED:
                return f"'{file_path}' exists and is committed"
            case CommitedFileStatus.NOT_COMMITTED:
                return f"'{file_path}' exists but is not committed"
            case CommitedFileStatus.MISSING:
                return f"'{file_path}' does not exist"
            case _:
                assert False


class CheckFileTask(Task):
    """
    This task checks if a file exists and is committed in a Git repository, or the inverse.
    """

    #: The path to the file to check. A relative path is considered relative to the project directory.
    file_to_check: Property[Path]

    #: The desired state of the file.
    state: Property[Literal["present", "absent"]] = Property.default("present")

    #: If set to True, the task will not fail if the file state does not match the desired state.
    #: Otherwise, the task will error.
    warn_only: Property[bool] = Property.default(False)

    def execute(self) -> TaskStatus:
        """Checks that a give file exists and has been committed to git."""
        status = check_file_status(self.project.directory / self.file_to_check.get())
        success: bool
        match status:
            case CommitedFileStatus.COMMITED:
                success = self.state.get() == "present"
            case CommitedFileStatus.NOT_COMMITTED | CommitedFileStatus.MISSING:
                success = self.state.get() != "present"
            case _:
                assert False, f"Unknown status: {status}"

        return TaskStatus(
            TaskStatusType.SUCCEEDED
            if success
            else (TaskStatusType.WARNING if self.warn_only.get() else TaskStatusType.FAILED),
            status.to_description(self.file_to_check.get()),
        )

    def get_description(self) -> str | None:
        if self.state.get() == "present":
            return f"Checks if the file '{self.file_to_check.get()}' is committed."
        else:
            return f"Checks if the file '{self.file_to_check.get()}' is not committed."



from __future__ import annotations

import contextlib
import shlex
import tempfile
from collections.abc import Sequence
from pathlib import Path

import deprecated
from kraken.common import flatten
from kraken.core import Project, Property

from kraken.std.docker.util.dockerapi import docker_load, docker_run
from kraken.std.docker.util.dockerfile import render_docker_auth, update_run_commands

from .base_build_task import BaseBuildTask


class KanikoBuildTask(BaseBuildTask):
    """
    An implementation for building Docker images with Kaniko.

    In order to make secrets available in the Dockerfile under `/run/secrets`, and `RUN` command that is executed
    by the root user will be surrounded by additional shell commands to link `/kaniko/secrets` to `/run/secrets`.
    This is to ensure a degree of compatibility with BuildKit, which mounts secrets to `/run/secrets` when using
    the `--secret` flag.

    However, this compatibility is limited because we can only actually create a at `/run/secrets` if the `RUN`
    command runs as root. Any command that is executed as a non-root user will not be able to access the secrets.
    We use a heuristic to detect this, but it is not perfect. If you need to use secrets in a `RUN` command that
    is executed as a non-root user, you will need to read from `/kaniko/secrets` instead.

    If you start from a base image that doesn't start as the root user, you will need to explicitly add a `USER`
    command to your Dockerfile to either inform the `KanikoBuildTask` that the subsequent commands are not run as
    root, or use it to switch to the root user. Otherwise, the `KanikoBuildTask` will assume that the subsequent
    commands are run as root and it can not know in advance if they are or not.
    """

    kaniko_image: Property[str] = Property.default("gcr.io/kaniko-project/executor:v1.9.0-debug")
    kaniko_context: Property[str] = Property.default("/workspace")
    kaniko_cache_copy_layers: Property[bool] = Property.default(True)
    kaniko_snapshot_mode: Property[str] = Property.default("redo")
    kaniko_secrets_mount_dir: Property[str] = Property.default("/kaniko/secrets")
    kaniko_secrets_from_env: Property[Sequence[str]] = Property.default(())
    kaniko_use_compressed_caching: Property[bool] = Property.default(True)

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.preprocess_dockerfile.set(True)

    def render_main_script(self, executor_command: list[str]) -> str:
        """Renders the shell script that will be executed in the Kaniko container."""

        docker_config = render_docker_auth(self.auth.get())

        script = []
        script += [
            "mkdir -p /kaniko/.docker",
            "cat << EOF > /kaniko/.docker/config.json",
            docker_config,
            "EOF",
        ]

        secrets_mount_dir = self.kaniko_secrets_mount_dir.get()
        if self.secrets.get():
            script += [f"mkdir -p {shlex.quote(secrets_mount_dir)}"]
            for secret, value in self.secrets.get().items():
                script += [f"echo {shlex.quote(value)} > {shlex.quote(secrets_mount_dir + '/' + secret)}"]
        if self.kaniko_secrets_from_env.get():
            for env in self.kaniko_secrets_from_env.get():
                script += [f"echo ${env} > {shlex.quote(secrets_mount_dir + '/' + env)}"]

        script += [" ".join(map(shlex.quote, executor_command))]
        return "\n".join(script)

    def get_kaniko_executor_command(self, dockerfile: str | None, tar_path: str | None) -> list[str]:
        if tar_path and not self.tags:
            raise ValueError("Need at least one destination (tag) when exporting to an image tarball")
        executor_command = ["/kaniko/executor"]
        executor_command += flatten(("--build-arg", f"{key}={value}") for key, value in self.build_args.get().items())
        cache_repo = self.cache_repo.get()
        if cache_repo:
            executor_command += ["--cache-repo", cache_repo]
        if self.cache.get_or(False):
            executor_command += ["--cache=true"]
        executor_command += flatten(("--destination", destination) for destination in self.tags.get())
        if dockerfile:
            executor_command += ["--dockerfile", dockerfile]
        if not self.push.get():
            executor_command += ["--no-push"]
        executor_command += ["--snapshotMode", self.kaniko_snapshot_mode.get()]
        if self.squash.get():
            executor_command += ["--single-snapshot"]
        if self.kaniko_cache_copy_layers.get():
            executor_command += ["--cache-copy-layers"]
        if not self.kaniko_use_compressed_caching.get():
            executor_command += ["--compressed-caching=false"]
        target = self.target.get()
        if target:
            executor_command += ["--target", target]
        if tar_path:
            executor_command += ["--tarPath", tar_path]
        executor_command += ["--context", self.kaniko_context.get()]
        return executor_command

    @deprecated.deprecated("use KanikoBuildTask.render_main_script() instead", version="0.3.6")
    def _render_main_script(self, executor_command: list[str]) -> str:
        return self.render_main_script(executor_command)

    @deprecated.deprecated("use KanikoBuildTask.get_kaniko_executor_command() instead", version="0.3.6")
    def _get_kaniko_executor_command(self, dockerfile: str | None, tar_path: str | None) -> list[str]:
        return self.get_kaniko_executor_command(dockerfile, tar_path)

    def _build(
        self,
        exit_stack: contextlib.ExitStack,
    ) -> None:
        volumes = [f"{self.build_context.get().absolute()}:{self.kaniko_context.get()}"]

        # If the Dockerfile is not relative to the build context, we need to mount it explicitly.
        dockerfile = self.dockerfile.get()
        in_container_dockerfile: str | None = None
        try:
            in_container_dockerfile = str(dockerfile.absolute().relative_to(self.build_context.get().absolute()))
        except ValueError:
            in_container_dockerfile = "/kaniko/Dockerfile"
            volumes += [f"{dockerfile.absolute()}:{in_container_dockerfile}"]

        # If the image needs to be loaded into the Docker daemon after building, we need to always
        # export it to a file.
        image_output_file = self.image_output_file.get_or(None)
        if self.load.get() and not image_output_file:
            tempdir = exit_stack.enter_context(tempfile.TemporaryDirectory())
            image_output_file = Path(tempdir) / "image.tgz"

        # Construct the tar path for inside the container.
        tar_path: str | None = None
        if image_output_file:
            volumes += [f"{image_output_file.parent.absolute()}:/kaniko/out"]
            tar_path = f"/kaniko/out/{image_output_file.name}"

        executor_command = self._get_kaniko_executor_command(in_container_dockerfile, tar_path)

        script = self._render_main_script(executor_command)

        result = docker_run(
            image=self.kaniko_image.get(),
            args=["sh", "-c", script],
            entrypoint="",
            remove=True,
            volumes=volumes,
            workdir=self.kaniko_context.get(),
            platform=self.platform.get_or(None),
        )

        if result != 0:
            raise Exception(f"Kaniko build failed with exit code {result}")

        if self.load.get():
            assert image_output_file is not None, "image_output_file is expected to be set when config.load == True"
            result = docker_load(image_output_file)
            if result != 0:
                raise Exception(f"Docker load failed with exit code {result}")

    # BaseBuildTask overrides

    def _preprocess_dockerfile(self, dockerfile: Path) -> str:
        return update_run_commands(
            dockerfile.read_text(),
            prefix="ln -sf /kaniko/secrets /run/secrets && ( ",
            suffix=" ); __ret=$?; unlink /run/secrets; exit $__ret",
            # We can only link to /run/secrets if the user running the command is the root user. In any other
            # case, we cannot link to /run/secrets and reading the secrets from /run/secrets in the Dockerfile
            # will fail.
            only_for_root_user=True,
        )

    # Task overrides

    def finalize(self) -> None:
        if self.cache.get() and not self.push.get() and not self.cache_repo.get():
            self.logger.warning(
                "Disabling cache in Kaniko build %s because it must be combined with push or cache_repo",
                self,
            )
            self.cache.set(False)
        cache_repo = self.cache_repo.get_or(None)
        if cache_repo and ":" in cache_repo:
            raise ValueError(f"Kaniko --cache-repo argument cannot contain `:` (got: {cache_repo!r})")
        return super().finalize()

    def execute(self) -> None:
        with contextlib.ExitStack() as exit_stack:
            self._build(exit_stack)



from __future__ import annotations

from pathlib import Path

from kraken.core import Property, TaskStatus

from ._cargo_sqlx import CargoBaseSqlxTask


class CargoSqlxMigrateTask(CargoBaseSqlxTask):
    """Apply SQL migrations using cargo sqlx. If the database URL is not provided, it will default to the environment
    variable DATABASE_URL. If the database does not exist, it will be created."""

    migrations: Property[Path]

    def execute(self) -> TaskStatus:
        result = self.db_create()
        if result.is_not_ok():
            return result

        result = self.migrate_run()
        if result.is_not_ok():
            return result

        return TaskStatus.succeeded()

    def db_create(self) -> TaskStatus:
        return self._execute_command(["db", "create"])

    def migrate_run(self) -> TaskStatus:
        arguments = ["migrate", "run"]
        if self.migrations.is_filled():
            arguments.extend(["--source", str(self.migrations.get().absolute())])

        return self._execute_command(arguments)

    def get_description(self) -> str | None:
        return "Apply SQL migrations using cargo sqlx"



from __future__ import annotations

from difflib import Differ
from pathlib import Path

from kraken.common.path import try_relative_to
from kraken.common.strings import as_string
from kraken.core import Property, Task, TaskStatus
from termcolor import colored


class CheckFileContentsTask(Task):
    """The CheckFileContentsTask will compare the contents of a file of a file with the content specified in the
    :attr:`content` property and error if it does not match. This is usually used in combination with a
    :class:`RenderFileTask`."""

    description = 'Check if "%(file)s" is up to date.'

    file: Property[Path]
    content: Property[str | bytes]
    encoding: Property[str]
    render_prepare: Property[TaskStatus]
    update_task_name: Property[str | None] = Property.default(None)
    show_diff: Property[bool] = Property.default(True)

    def _show_diff(self, a: str, b: str) -> None:
        differ = Differ()
        for line in differ.compare(a.splitlines(keepends=True), b.splitlines(keepends=True)):
            print(line, end="")

    # Task

    def prepare(self) -> TaskStatus:
        return self.render_prepare.get()

    def execute(self) -> TaskStatus | None:
        file = try_relative_to(self.project.directory / self.file.get())
        file_fmt = colored(str(file), "yellow", attrs=["bold"])

        if update_task_name := self.update_task_name.get():
            uptask = colored(update_task_name, "blue", attrs=["bold"])
            message_suffix = f", run {uptask} to generate it"
        else:
            message_suffix = ""

        if not file.exists():
            return TaskStatus.failed(f'file "{file_fmt}" does not exist{message_suffix}')
        if not file.is_file():
            return TaskStatus.failed(f'"{file}" is not a file')

        encoding = self.encoding.get()
        if (file_content := file.read_text(encoding)) != (content := as_string(self.content.get(), encoding)):
            if self.show_diff.get():
                self._show_diff(file_content, content)
            return TaskStatus.failed(f'file "{file_fmt}" is not up to date{message_suffix}')

        return TaskStatus.succeeded(f'file "{file_fmt}" is up to date')



from kraken.std.python.buildsystem.poetry import PoetryPyprojectHandler
from kraken.std.python.pyproject import PackageIndex, Pyproject

EXAMPLE_POETRY_PYPROJECT = """
[tool.poetry]
name = "poetry-project"
version = "0.1.0"

[[tool.poetry.source]]
name = "foo"
url = "https://foo.bar/simple/"
priority = "supplemental"
"""

EXAMPLE_POETRY_PYPROJECT_WITH_LEGACY_SOURCE_CONFIG = """
[[tool.poetry.source]]
name = "foo"
url = "https://foo.bar/simple/"
default = true

[[tool.poetry.source]]
name = "bar"
url = "https://bar.com/simple/"
secondary = true
"""


def test__PoetryPyprojectHandler__getters() -> None:
    handler = PoetryPyprojectHandler(Pyproject.read_string(EXAMPLE_POETRY_PYPROJECT))
    assert handler.get_name() == "poetry-project"
    assert handler.get_version() == "0.1.0"


def test__PoetryPyprojectHandler__set_version() -> None:
    handler = PoetryPyprojectHandler(Pyproject.read_string(EXAMPLE_POETRY_PYPROJECT))
    handler.set_version("2.0.0")
    assert handler.raw["tool"]["poetry"]["version"] == "2.0.0"


def test__PoetryPyprojectHandler__get_package_indexes() -> None:
    handler = PoetryPyprojectHandler(Pyproject.read_string(EXAMPLE_POETRY_PYPROJECT))
    assert handler.get_package_indexes() == [
        PackageIndex(
            alias="foo",
            index_url="https://foo.bar/simple/",
            priority=PackageIndex.Priority.supplemental,
            verify_ssl=True,
        )
    ]


def test__PoetryPyprojectHandler__get_package_indexes__with_legacy_source_config() -> None:
    handler = PoetryPyprojectHandler(Pyproject.read_string(EXAMPLE_POETRY_PYPROJECT_WITH_LEGACY_SOURCE_CONFIG))
    assert handler.get_package_indexes() == [
        PackageIndex(
            alias="foo",
            index_url="https://foo.bar/simple/",
            priority=PackageIndex.Priority.default,
            verify_ssl=True,
        ),
        PackageIndex(
            alias="bar",
            index_url="https://bar.com/simple/",
            priority=PackageIndex.Priority.secondary,
            verify_ssl=True,
        ),
    ]


def test__PoetryPyprojectHandler__set_package_indexes__to_empty_list() -> None:
    handler = PoetryPyprojectHandler(Pyproject.read_string(EXAMPLE_POETRY_PYPROJECT))
    handler.set_package_indexes([])
    assert handler.raw["tool"]["poetry"]["source"] == []


def test__PoetryPyprojectHandler__set_package_indexes__to_various_indexes() -> None:
    handler = PoetryPyprojectHandler(Pyproject.read_string(EXAMPLE_POETRY_PYPROJECT))
    handler.set_package_indexes(
        [
            PackageIndex("a", "https://a.com", PackageIndex.Priority.primary, True),
            PackageIndex("c", "https://c.com", PackageIndex.Priority.supplemental, False),
            PackageIndex("b", "https://b.com", PackageIndex.Priority.default, True),
        ]
    )
    assert handler.raw["tool"]["poetry"]["source"] == [
        {"name": "a", "url": "https://a.com", "priority": "primary"},
        {"name": "c", "url": "https://c.com", "priority": "supplemental"},
        {"name": "b", "url": "https://b.com", "priority": "default"},
    ]



from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path

from kraken.common import Supplier
from kraken.core import Project, Property

from .base_task import EnvironmentAwareDispatchTask


class PylintTask(EnvironmentAwareDispatchTask):
    description = "Lint Python source files with Pylint"
    python_dependencies = ["pylint"]

    config_file: Property[Path]
    additional_args: Property[Sequence[str]] = Property.default_factory(list)

    # EnvironmentAwareDispatchTask

    def get_execute_command(self) -> list[str]:
        command = ["pylint", str(self.settings.source_directory)] + self.settings.get_tests_directory_as_args()
        if self.config_file.is_filled():
            command += ["--rcfile", str(self.config_file.get())]
        command += self.additional_args.get()
        return command


def pylint(
    *,
    name: str = "python.pylint",
    project: Project | None = None,
    config_file: Path | Supplier[Path] | None = None,
    additional_args: Sequence[str] | Property[Sequence[str]] = (),
) -> PylintTask:
    project = project or Project.current()
    task = project.task(name, PylintTask, group="lint")
    task.config_file = config_file
    task.additional_args = additional_args
    return task



from __future__ import annotations

from pathlib import Path

from kraken.core import Project, Property, TaskStatus

from kraken.std.python.buildsystem import PythonBuildSystem
from kraken.std.util.render_file_task import RenderFileTask

from ..pyproject import Pyproject
from ..settings import PythonSettings, python_settings


class UpdatePyprojectTask(RenderFileTask):
    """Updates the `pyproject.toml` to ensure certain configuration options are set."""

    build_system: Property[PythonBuildSystem]
    settings: Property[PythonSettings]

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.content.setcallable(lambda: self.get_file_contents(self.file.get()))

    def get_file_contents(self, file: Path) -> str:
        pyproject = Pyproject.read(file)
        settings = self.settings.get()
        assert settings.build_system
        settings.build_system.update_pyproject(settings, pyproject)
        return pyproject.to_toml_string()

    def prepare(self) -> TaskStatus:
        settings = self.settings.get()
        if not settings.build_system:
            return TaskStatus.skipped("no build system")
        return super().prepare()


def update_pyproject_task(
    *,
    name: str = "pyproject.update",
    group: str = "apply",
    project: Project | None = None,
) -> UpdatePyprojectTask:
    project = project or Project.current()
    task = project.task(name, UpdatePyprojectTask, group=group)
    task.settings = python_settings(project)
    task.file = project.directory / "pyproject.toml"
    task.create_check()
    return task



from pathlib import Path

from kraken.common.supplier import Supplier
from kraken.core import Project, Property, Task

from kraken.std.util.render_file_task import RenderFileTask, render_file


class BaseBuildTask(Task):
    """Base class for tasks that build Docker images. Subclasses implement converting the task properties into
    the invokation for a Docker build backend."""

    build_context: Property[Path]
    dockerfile: Property[Path] = Property.default(Path("Dockerfile"))
    auth: Property[dict[str, tuple[str, str]]] = Property.default_factory(dict)
    platform: Property[str]
    build_args: Property[dict[str, str]] = Property.default_factory(dict)
    secrets: Property[dict[str, str]] = Property.default_factory(dict)
    cache_repo: Property[str | None] = Property.default(None)
    cache: Property[bool] = Property.default(True)
    tags: Property[list[str]] = Property.default_factory(list)
    push: Property[bool] = Property.default(False)
    squash: Property[bool] = Property.default(False)
    target: Property[str | None] = Property.default(None)
    image_output_file: Property[Path | None] = Property.default(None)
    load: Property[bool] = Property.default(False)

    # If enabled, a separate Dockerfile preprocessing task will be created during finalize().
    # Implementations that typically require processing will enable this property by default.
    preprocess_dockerfile: Property[bool] = Property.default(False)

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.build_context.set(project.directory)
        self.preprocessor_task: RenderFileTask | None = None

    def _preprocess_dockerfile(self, dockerfile: Path) -> str:
        """Internal. Can be implemented by subclasses. Preprocess the Dockerfile."""

        return dockerfile.read_text()

    def create_preprocessor_task(self, group: str | None = None) -> RenderFileTask:
        assert self.preprocess_dockerfile.get(), f"no preprocessing requested: {self}"
        assert not self.preprocessor_task, f"preprocessor task already exists: {self.preprocessor_task}"
        tempfile = self.project.build_directory / self.name / "Dockerfile"
        dockerfile = self.dockerfile.value
        task = render_file(
            name=self.name + ".preprocess",
            description="Preprocess the Dockerfile.",
            create_check=False,
            group=group,
            file=tempfile,
            content=Supplier.of_callable(lambda: self._preprocess_dockerfile(dockerfile.get()), [dockerfile]),
            project=self.project,
        )[0]
        self.dockerfile.set(task.file)
        self.preprocessor_task = task
        return task

    # Task

    def finalize(self) -> None:
        if not self.preprocessor_task and self.preprocess_dockerfile.get():
            self.create_preprocessor_task()
        return super().finalize()



from __future__ import annotations

from kraken.core import Property

from .cargo_build_task import CargoBuildTask


class CargoClippyTask(CargoBuildTask):
    """Runs `cargo clippy` for linting or applying suggestions."""

    fix: Property[bool] = Property.default(False)
    allow: Property[str | None] = Property.default("staged")

    # CargoBuildTask

    def get_cargo_command(self, env: dict[str, str]) -> list[str]:
        command = ["cargo", "clippy"]
        if self.fix.get():
            command += ["--fix"]
            allow = self.allow.get()
            if allow == "staged":
                command += ["--allow-staged"]
            elif allow == "dirty":
                command += ["--allow-dirty", "--allow-staged"]
            elif allow is not None:
                raise ValueError(f"invalid allow: {allow!r}")
        return command



from __future__ import annotations

from kraken.core import Project, Property, Task, TaskStatus

from ..settings import PythonSettings, python_settings


class LoginTask(Task):
    settings: Property[PythonSettings]

    def prepare(self) -> TaskStatus | None:
        settings = self.settings.get()
        if not settings.build_system:
            return TaskStatus.skipped("no build system configured")
        if not settings.build_system.requires_login():
            return TaskStatus.skipped("build system requires no log in")
        return TaskStatus.pending()

    def execute(self) -> None:
        settings = self.settings.get()
        assert settings.build_system is not None
        settings.build_system.login(settings)


def login(*, name: str = "python.login", project: Project | None = None) -> LoginTask:
    project = project or Project.current()
    task = project.task(name, LoginTask)
    task.settings = python_settings(project)
    return task



"""Python build system helper functions """

from __future__ import annotations

import re
from pathlib import Path


def update_python_version_str_in_source_files(as_version: str, package_directory: Path) -> int:
    """
    Updates the `__version__` in Python source files in the given *package_directory*. The constant is commonly
    contained in the files `__init__.py`, `__about__.py` and `_version.py`. No such files in subdirectories will
    be updated. The *package_directory* is expected to point at the directory that contains directly the file(s)
    with the `__version__` constant.

    Returns the number of version references that have been replaced.
    """

    FILENAMES = ["__init__.py", "__about__.py", "_version.py"]
    VERSION_REGEX = r'__version__\s*=\s*[\'"]([^\'"]+)[\'"]'
    replace_with = f'__version__ = "{as_version}"'
    sum_replaced = 0

    for source_file_path in package_directory.glob("*.py"):
        if source_file_path.name not in FILENAMES:
            continue
        source_file_path_ = Path(source_file_path)
        content = source_file_path_.read_text()
        (content, n_replaced) = re.subn(VERSION_REGEX, replace_with, content, flags=re.M)
        if n_replaced > 0:
            source_file_path_.write_text(content)
        sum_replaced += n_replaced

    return sum_replaced



from __future__ import annotations

import dataclasses
from collections.abc import Sequence
from pathlib import Path

from kraken.common import Supplier
from kraken.core import Project, Property

from .base_task import EnvironmentAwareDispatchTask


class IsortTask(EnvironmentAwareDispatchTask):
    python_dependencies = ["isort"]

    check_only: Property[bool] = Property.default(False)
    config_file: Property[Path]
    additional_files: Property[Sequence[Path]] = Property.default_factory(list)

    # EnvironmentAwareDispatchTask

    def get_execute_command(self) -> list[str]:
        command = ["isort", str(self.settings.source_directory)] + self.settings.get_tests_directory_as_args()
        command += [str(p) for p in self.additional_files.get()]
        if self.check_only.get():
            command += ["--check-only", "--diff"]
        if self.config_file.is_filled():
            command += ["--settings-file", str(self.config_file.get().absolute())]
        return command

    # Task

    def get_description(self) -> str | None:
        if self.check_only.get():
            return "Check Python source files formatting with isort."
        else:
            return "Format Python source files with isort."


@dataclasses.dataclass
class IsortTasks:
    check: IsortTask
    format: IsortTask


def isort(
    *,
    name: str = "python.isort",
    project: Project | None = None,
    config_file: Path | Supplier[Path] | None = None,
    additional_files: Sequence[Path] | Supplier[Sequence[Path]] = (),
) -> IsortTasks:
    # TODO (@NiklasRosenstein): We may need to ensure an order to isort and block somehow, sometimes they yield
    #       slightly different results based on the order they run.
    project = project or Project.current()

    check_task = project.task(f"{name}.check", IsortTask, group="lint")
    check_task.check_only = True
    check_task.config_file = config_file
    check_task.additional_files = additional_files

    format_task = project.task(name, IsortTask, group="fmt")
    format_task.check_only = False
    format_task.config_file = config_file
    format_task.additional_files = additional_files

    return IsortTasks(check_task, format_task)



from __future__ import annotations

import dataclasses
import enum
import re
import subprocess as sp
from pathlib import Path


def git_describe(path: Path | None, tags: bool = True, dirty: bool = True) -> str:
    """Describe a repository with tags.

    :param path: The directory in which to describe.
    :param tags: Whether to include tags (adds the `--tags` flag).
    :param dirty: Whether to include if the directory tree is dirty (adds the `--dirty` flag).
    :raise ValueError: If `git describe` failed.
    :return: The Git head description.
    """

    command = ["git", "describe"]
    if tags:
        command.append("--tags")
    if dirty:
        command.append("--dirty")
    try:
        return sp.check_output(command, cwd=path).decode().strip()
    except sp.CalledProcessError:
        count = int(sp.check_output(["git", "rev-list", "HEAD", "--count"], cwd=path).decode().strip())
        short_rev = sp.check_output(["git", "rev-parse", "--short", "HEAD"], cwd=path).decode().strip()
        return f"0.0.0-{count}-g{short_rev}"


@dataclasses.dataclass
class GitVersion:
    """Represents a "git version" that has a major, minor and patch version and optionally a commit distance."""

    @dataclasses.dataclass
    class PreRelease:
        @enum.unique
        class Kind(str, enum.Enum):
            ALPHA = "alpha"
            BETA = "beta"
            RC = "rc"

        kind: Kind
        value: int

    @dataclasses.dataclass
    class CommitDistance:
        value: int
        sha: str

    major: int
    minor: int
    patch: int
    pre_release: PreRelease | None
    distance: CommitDistance | None
    dirty: bool

    @staticmethod
    def parse(value: str) -> GitVersion:
        GIT_VERSION_REGEX = r"^(\d+)\.(\d+)\.(\d+)(?:-(alpha|beta|rc).(\d+))?(?:-(\d+)-g(\w+))?(-dirty)?$"
        match = re.match(GIT_VERSION_REGEX, value)
        if not match:
            raise ValueError(f"not a valid GitVersion: {value!r}")
        if match.group(4):
            pre_release = GitVersion.PreRelease(
                kind=GitVersion.PreRelease.Kind(match.group(4)), value=int(match.group(5))
            )
        else:
            pre_release = None
        if match.group(6):
            distance = GitVersion.CommitDistance(value=int(match.group(6)), sha=match.group(7))
        else:
            distance = None
        return GitVersion(
            major=int(match.group(1)),
            minor=int(match.group(2)),
            patch=int(match.group(3)),
            pre_release=pre_release,
            distance=distance,
            dirty=match.group(8) is not None,
        )

    def format(self, distance: bool = True, sha: bool = True, dirty: bool = False) -> str:
        result = f"{self.major}.{self.minor}.{self.patch}"
        if self.pre_release:
            result = f"{result}-{self.pre_release.kind.value}.{self.pre_release.value}"
        if self.distance and distance:
            result = f"{result}-{self.distance.value}"
            if sha:
                result = f"{result}-g{self.distance.sha}"
            if self.dirty and dirty:
                result = f"{result}-dirty"
        return result



from __future__ import annotations

from pathlib import Path

from kraken.common import Supplier
from kraken.core import Project, Property

from .base_task import EnvironmentAwareDispatchTask


class Flake8Task(EnvironmentAwareDispatchTask):
    """
    Lint Python source files with Flake8.
    """

    description = "Lint Python source files with Flake8."
    python_dependencies = ["flake8"]

    config_file: Property[Path]
    additional_args: Property[list[str]] = Property.default_factory(list)

    # EnvironmentAwareDispatchTask

    def get_execute_command(self) -> list[str]:
        command = ["flake8", str(self.settings.source_directory)] + self.settings.get_tests_directory_as_args()
        if self.config_file.is_filled():
            command += ["--config", str(self.config_file.get().absolute())]
        command += self.additional_args.get()
        return command


def flake8(
    *, name: str = "python.flake8", project: Project | None = None, config_file: Path | Supplier[Path] | None = None
) -> Flake8Task:
    project = project or Project.current()
    task = project.task(name, Flake8Task, group="lint")
    if config_file is not None:
        task.config_file = config_file
    return task



from kraken.std.git.gitignore.generated import join_generated_section, split_generated_section
from kraken.std.git.gitignore.parser import GitignoreFile

LEGACY_GITIGNORE = """
### START-GENERATED-CONTENT [HASH: 3d9a3142ccc0b9baed88a8427ff555e32ebc44d88e851c24d35a8e6ea7901ad3]
# -------------------------------------------------------------------------------------------------
# THIS SECTION WAS AUTOMATICALLY GENERATED BY KRAKEN; DO NOT MODIFY OR YOUR CHANGES WILL BE LOST.
# If you need to define custom gitignore rules, add them below
# -------------------------------------------------------------------------------------------------
### [PARAMETERS_HASH: 5a7dde844945ce553190de65b75086d9f6f05eec8cabc8c52cd0b202285a91a1]

# abc
foobar
# -------------------------------------------------------------------------------------------------
### END-GENERATED-CONTENT

custom
"""

NEW_GITIGNORE = """
# GENERATED-CONTENT-START
# Do not edit
foo
bar
# GENERATED-CONTENT-END
"""


def test__split_generated_section__can_parse_legacy_gitignore() -> None:
    user1, generated, user2 = split_generated_section(GitignoreFile.parse(LEGACY_GITIGNORE))
    assert user1 == [""]
    assert generated == [
        "",
        "# abc",
        "foobar",
    ]
    assert user2 == [
        "",
        "custom",
    ]


def test__split_generated_section__can_parse_new_gitignore() -> None:
    user1, generated, user2 = split_generated_section(GitignoreFile.parse(NEW_GITIGNORE))
    assert user1 == [""]
    assert generated == [
        "# Do not edit",
        "foo",
        "bar",
    ]
    assert user2 == []


def test__join_generated_section__same_as_parsed() -> None:
    user1, generated, user2 = split_generated_section(GitignoreFile.parse(NEW_GITIGNORE))
    assert NEW_GITIGNORE == "\n".join(join_generated_section(user1, generated, user2)) + "\n"



from __future__ import annotations

from pathlib import Path

from kraken.common import Supplier
from kraken.core import Project, Property, Task, TaskStatus

from ..buildsystem import PythonBuildSystem
from ..pyproject import Pyproject
from ..settings import PythonSettings, python_settings


class UpdateLockfileTask(Task):
    settings: Property[PythonSettings]
    build_system: Property[PythonBuildSystem | None]
    pyproject_toml: Property[Path]

    def get_description(self) -> str | None:
        build_system = self.build_system.get()
        return (
            f"Update dependencies in your Python project. [build system: "
            f"{build_system.name if build_system else None}]"
        )

    def execute(self) -> TaskStatus:
        build_system = self.build_system.get()
        if not build_system:
            return TaskStatus.failed("no build system configured")
        settings = self.settings.get()
        pyproject = Pyproject.read(self.pyproject_toml.get())
        return build_system.update_lockfile(settings, pyproject)


def update_lockfile_task(
    *,
    name: str = "python.update",
    group: str | None = "update",
    project: Project | None = None,
) -> UpdateLockfileTask:
    """Creates an update task for the given project.

    The update task relies on the build system configured in the Python project settings."""

    project = project or Project.current()
    task = project.task(name, UpdateLockfileTask, group=group)
    task.settings = python_settings(project)
    task.build_system = Supplier.of_callable(lambda: python_settings(project).build_system)
    task.pyproject_toml = project.directory / "pyproject.toml"
    return task



"""
This module provides a thin client for the gitignore.io API to retrieve template `.gitignore` files.

Run this module to update the `gitignore-io-tokens.json.gz` file.
"""

import logging
from collections.abc import Sequence
from functools import lru_cache
from gzip import compress, decompress
from json import dumps, loads
from pathlib import Path
from typing import cast

from kraken.std import http

logger = logging.getLogger(__name__)


GITIGNORE_API_URL = "https://www.toptal.com/developers/gitignore/api/"

# When running this script, these tokens will be fetched and written into a local cache file.
# When these tokens are requested, we don't need to make a request to gitignore.io at build time.
TOKENS = [
    # Platforms
    "linux",
    "macos",
    "windows",
    # IDEs
    "clion",
    "emacs",
    "intellij",
    "jupyternotebooks",
    "pycharm",
    "vim",
    "visualstudiocode",
    # Tooling
    "gcov",
    "git",
    "yarn",
    # Languages
    "c",
    "c++",
    "matlab",
    "node",
    "python",
    "react",
    "rust-analyzer",
    "rust",
]

# The file where we store the cached token data.
OUTPUT_FILE = Path(__file__).parent / "data" / "gitignore-io-tokens.json.gz"


def gitignore_io_fetch(tokens: Sequence[str]) -> str:
    """
    Fetch a `.gitignore` file from the gitignore.io API.
    """

    url = GITIGNORE_API_URL + ",".join(tokens)
    response = http.get(url)
    response.raise_for_status()
    return response.text.replace("\r\n", "\n")


def gitignore_io_fetch_cached(tokens: Sequence[str], backfill: bool) -> str:
    """
    Compile the `.gitignore` file for the given tokens from the cache. Any tokens that are not
    cached by be backfilled by making another request to gitignore.io if *backfill* is enabled.
    If *backfill* is disabled and a token is not cached, an #ValueError will be raised.
    """

    cache = load_token_cache()
    not_cached = set(tokens) - set(cache.keys())
    if not_cached and not backfill:
        raise ValueError(
            f"The following gitignore.io tokens are not distributed as part of kraken-std: {', '.join(not_cached)}"
            "\nBackfill is disabled, so this error is raised instead of making another HTTP request to "
            "gitignore.io."
        )

    additional_tokens = ""
    if not_cached:
        additional_tokens = gitignore_io_fetch([t for t in tokens if t in not_cached])  # Consistent order

    return "\n".join([cache[t] for t in tokens if t in cache] + [additional_tokens])


def write_token_cache(tokens: dict[str, str]) -> None:
    # NOTE(@NiklasRosenstein): We set mtime to 0 to ensure the Gzip compression result is stable.
    OUTPUT_FILE.write_bytes(compress(dumps(tokens).encode("utf-8"), mtime=0))


@lru_cache
def load_token_cache() -> dict[str, str]:
    if not OUTPUT_FILE.is_file():
        logger.warning("No gitignore.io token cache found at %s", OUTPUT_FILE)
        return {}
    return cast(dict[str, str], loads(decompress(OUTPUT_FILE.read_bytes()).decode("utf-8")))


def main() -> None:
    logging.basicConfig(level=logging.INFO)

    result = {}
    for token in sorted(TOKENS):
        logger.info("Fetching token %s", token)
        result[token] = gitignore_io_fetch([token])

    write_token_cache(result)


if __name__ == "__main__":
    main()



import subprocess
from collections.abc import Sequence
from pathlib import Path

from kraken.core import Property, Task, TaskStatus


class CargoDenyTask(Task):
    description = "Executes cargo deny to verify dependencies."
    checks: Property[Sequence[str]] = Property.default_factory(list)
    config_file: Property[Path]
    error_message: Property[str | None] = Property.default(None)

    def execute(self) -> TaskStatus:
        command = ["cargo", "deny", "check"]

        if self.config_file.is_filled():
            command.extend(["--config", str(self.config_file.get().absolute())])

        command.extend(self.checks.get())

        result = subprocess.run(command, cwd=self.project.directory)
        if result.returncode == 0:
            return TaskStatus.succeeded()

        return self.error_message.map(
            lambda message: TaskStatus.failed(message)
            if message is not None
            else TaskStatus.from_exit_code(command, result.returncode)
        ).get()



# A linter that ensures every HTTP request goes through this 'http' module, instead of using httpx or requests directly

from __future__ import annotations

import ast
from collections.abc import Generator

from . import __all__ as WRAPPED_HTTP_FUNCTIONS

HTTPX_ERROR_CODE = "KRA001"

PLAIN_HTTPX_MESSAGE = f"""{HTTPX_ERROR_CODE} using `httpx` directly is disallowed, because it may not play well with
corporate proxies.
Use kraken.std.http instead (which wraps calls to httpx).

See https://github.com/kraken-build/kraken-std/pull/132
"""

REQUESTS_ERROR_CODE = "KRA002"

PLAIN_REQUESTS_MESSAGE = f"""{REQUESTS_ERROR_CODE} using `requests` directly is disallowed, because it may not play well
with corporate proxies (unless `pip-system-certs` is installed).
Use kraken.std.http instead (which wraps calls to httpx).

See https://github.com/kraken-build/kraken-std/pull/132
"""


class BareCallsVisitor(ast.NodeVisitor):
    def __init__(self, disallowed_module: str) -> None:
        self.bare_calls: list[ast.AST] = []
        self.disallowed_module = disallowed_module

    def visit_Call(self, node: ast.Call) -> None:
        if BareCallsVisitor.call_is_bare(node, self.disallowed_module):
            self.bare_calls.append(node)

    @classmethod
    def call_is_bare(cls, call: ast.Call, disallowed_module: str) -> bool:
        """
        We don't want to forbid 'import httpx', because we may still want to use e.g. error codes from there
        Let's only forbid calls to functions that are wrapped by our http module
        """
        fn = call.func
        return (
            getattr(fn, "attr", None) in WRAPPED_HTTP_FUNCTIONS
            and hasattr(fn, "value")
            and BareCallsVisitor.call_is_from_disallowed_module(fn.value, disallowed_module)
        )

    @classmethod
    def call_is_from_disallowed_module(cls, fn_value: ast.expr, disallowed_module: str) -> bool:
        if getattr(fn_value, "attr", None) == disallowed_module:
            return hasattr(fn_value, "value") and BareCallsVisitor.call_is_from_disallowed_module(
                fn_value.value, disallowed_module
            )

        return getattr(fn_value, "id", None) == disallowed_module


class BanBareCalls:
    """
    Detects occurences of bare httpx or requests function calls.
    These are discouraged, as using the corresponding calls from the kraken.std.http usually work better
    in corporate networks.
    """

    def __init__(
        self, tree: ast.AST, filename: str, disallowed_module: str, error_message: str, final_type: type
    ) -> None:
        self.tree = tree
        self.disallowed_module = disallowed_module
        self.error_message = error_message
        self.final_type = final_type

    def run(self) -> Generator[tuple[int, int, str, type[BanBareHttpxCalls | BanBareRequestsCalls]], None, None]:
        visitor = BareCallsVisitor(self.disallowed_module)
        visitor.visit(self.tree)

        for node in visitor.bare_calls:
            yield node.lineno, node.col_offset, self.error_message, self.final_type


class BanBareHttpxCalls(BanBareCalls):
    def __init__(self, tree: ast.AST, filename: str) -> None:
        super().__init__(tree, filename, "httpx", PLAIN_HTTPX_MESSAGE, type(self))


class BanBareRequestsCalls(BanBareCalls):
    def __init__(self, tree: ast.AST, filename: str) -> None:
        super().__init__(tree, filename, "requests", PLAIN_REQUESTS_MESSAGE, type(self))



from __future__ import annotations

import logging
from abc import ABC
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, ClassVar, TypeAlias

import tomli
import tomli_w

logger = logging.getLogger(__name__)


class _PackageIndexPriority(str, Enum):
    """
    Poetry has a very granular representation of priorities for indices, so we inherit that. The priority
    should to be interpreted in the spirit of its definition in other tools.

    https://python-poetry.org/docs/repositories/#project-configuration
    """

    default = "default"
    primary = "primary"
    secondary = "secondary"  # Do not use
    supplemental = "supplemental"


@dataclass
class PackageIndex:
    """
    Represents a Python package index. Some tool-specific representations of package indices may not support
    all fields, in which case the implementation should trigger a warning or error if the alternative behaviour
    conflicts with the field value.
    """

    Priority: ClassVar[TypeAlias] = _PackageIndexPriority

    #: A name for the package index.
    alias: str

    #: The URL to find the packages at.
    index_url: str

    #: The priority of the index. Not all tools support it exactly how we model it in here (e.g. like Poetry),
    #: so the corresponding #PyprojectHandler implementation may need to do some translation.
    priority: Priority

    #: Whether SSL should be verified when connecting to the index. Not all tools support this.
    verify_ssl: bool


@dataclass
class Pyproject(dict[str, Any]):
    """
    Represents a raw `pyproject.toml` file in deserialized form.
    """

    path: Path | None

    def __init__(self, path: Path | None, data: Mapping[str, Any]) -> None:
        super().__init__(data)
        self.path = path

    @classmethod
    def read_string(cls, text: str) -> Pyproject:
        return cls(None, tomli.loads(text))

    @classmethod
    def read(cls, path: Path) -> Pyproject:
        with path.open("rb") as fp:
            return cls(path, tomli.load(fp))

    def save(self, path: Path | None = None) -> None:
        path = path or self.path
        if not path:
            raise RuntimeError("No path to save to")
        with path.open("wb") as fp:
            tomli_w.dump(self, fp)

    def to_toml_string(self) -> str:
        return tomli_w.dumps(self)


class PyprojectHandler(ABC):
    """
    A wrapper for a raw Pyproject to implement common read and mutation operations.
    """

    raw: Pyproject

    def __init__(self, raw: Pyproject) -> None:
        self.raw = raw

    def get_name(self) -> str | None:
        """
        Returns the current project name.

        This returns the name from the `project` section of the Pyproject.toml file as per PEP 621 [1].

        [1]: https://peps.python.org/pep-0621/#name
        """

        return self.raw.get("project", {}).get("name")  # type: ignore[no-any-return]

    def get_python_version_constraint(self) -> str | None:
        """
        Returns the current Python version constraint.

        This returns the Python version constraint from the `project` section of the Pyproject.toml file as per
        PEP 621 [1].

        [1]: https://peps.python.org/pep-0621/#requires-python
        """

        return self.raw.get("project", {}).get("requires-python")  # type: ignore[no-any-return]

    def get_version(self) -> str | None:
        """
        Returns the current project version. Returns #None if the project version can not be determined.

        This returns the version from the `project` section of the Pyproject.toml file as per PEP 621 [1].

        [1]: https://peps.python.org/pep-0621/#version
        """

        return self.raw.get("project", {}).get("version")  # type: ignore[no-any-return]

    def set_version(self, version: str | None) -> None:
        """
        Set (or unset) the project version.

        This sets the version from the `project` section of the Pyproject.toml file as per PEP 621 [1].

        [1]: https://peps.python.org/pep-0621/#version
        """

        project: dict[str, Any] | None = self.raw.get("project")
        if project is None:
            if version is None:
                return
            project = {"version": version}
            self.raw["project"] = project
        else:
            if version is None:
                project.pop("version", None)
            else:
                project["version"] = version

    def get_package_indexes(self) -> list[PackageIndex]:
        raise NotImplementedError("%s.get_package_indexes()" % type(self).__name__)

    def set_package_indexes(self, indexes: Sequence[PackageIndex]) -> None:
        raise NotImplementedError("%s.set_package_indexes()" % type(self).__name__)

    def set_path_dependencies_to_version(self, version: str) -> None:
        """
        Update all dependencies in the Pyproject that are path dependencies to the given version instead
        (such that they are no longer path dependencies).
        """

        raise NotImplementedError("%s.set_path_dependencies_to_version()" % type(self).__name__)



from __future__ import annotations

import subprocess as sp

from kraken.core import Task, TaskStatus


class CargoUpdateTask(Task):
    def execute(self) -> TaskStatus:
        command = ["cargo", "update"]
        return TaskStatus.from_exit_code(command, sp.call(command, cwd=self.project.directory))

    def get_description(self) -> str | None:
        return "Run `cargo update`."



import subprocess
from pathlib import Path

from kraken.core import Project, Property, Task, TaskStatus
from termcolor import colored

from kraken.std.python.settings import python_settings

from ..buildsystem import PythonBuildSystem


class InfoTask(Task):
    build_system: Property[PythonBuildSystem]

    def get_description(self) -> str:
        return "Displays metadata about the Kraken-managed environment."

    def execute(self) -> TaskStatus:
        """Displays metadata about the Kraken-managed environment."""
        python_path = self.get_python_path()
        virtual_env_path = self.get_virtual_env_path()
        try:
            version = self.get_python_version()
        except subprocess.CalledProcessError as error:
            return TaskStatus.failed(f"Error while getting the version of the current Python interpreter: {error}")

        print(
            colored(f" ---------- {self.build_system.get().name}-managed environment information ----------", "magenta")
        )
        print(colored("Python version:           ", "cyan"), colored(f"{version.strip()}", "blue"))
        print(colored("Python path:              ", "cyan"), colored(f"{python_path}", "blue"))
        print(colored("Virtual environment path: ", "cyan"), colored(f"{virtual_env_path}", "blue"))
        print(colored(" ------------------------------------------------------------", "magenta"))

        return TaskStatus.succeeded()

    def get_virtual_env_path(self) -> Path:
        """
        Returns the current virtual environment path
        """
        return self.build_system.get().get_managed_environment().get_path()

    def get_python_path(self) -> Path:
        """Returns the path of the Python interpreter of the Kraken-managed environment."""
        return self.get_virtual_env_path() / "bin" / "python"

    def get_python_version(self) -> str:
        """Returns the version of the Python interpreter of the Kraken-managed environment."""
        return subprocess.run(
            [self.get_python_path(), "--version"], stdout=subprocess.PIPE, shell=False, check=True
        ).stdout.decode("utf-8")


def info(*, project: Project | None = None, build_system: PythonBuildSystem | None = None) -> InfoTask:
    """
    This task displays a list of useful info on current Python and virtual environment settings.
    """

    project = project or Project.current()
    if build_system is None:
        build_system = python_settings().build_system

    task = project.task("python.info", InfoTask)
    task.build_system = build_system
    return task



from __future__ import annotations

import dataclasses
from collections.abc import Collection, Iterable, Sequence
from difflib import unified_diff
from pathlib import Path
from sys import stdout
from tempfile import TemporaryDirectory

from kraken.core import Project, Property, TaskStatus

from .. import python_settings
from .base_task import EnvironmentAwareDispatchTask


class PyUpgradeTask(EnvironmentAwareDispatchTask):
    description = "Upgrades to newer Python syntax sugars with pyupgrade."
    python_dependencies = ["pyupgrade"]

    keep_runtime_typing: Property[bool] = Property.default(False)
    additional_files: Property[Sequence[Path]] = Property.default_factory(list)
    python_version: Property[str]

    # EnvironmentAwareDispatchTask

    def get_execute_command(self) -> list[str]:
        return self.run_pyupgrade(self.additional_files.get(), ("--exit-zero-even-if-changed",))

    def run_pyupgrade(self, files: Iterable[Path], extra: Iterable[str]) -> list[str]:
        command = ["pyupgrade", f"--py{self.python_version.get_or('3').replace('.', '')}-plus", *extra]
        if self.keep_runtime_typing.get():
            command.append("--keep-runtime-typing")
        command.extend(str(f) for f in files)
        return command


class PyUpgradeCheckTask(PyUpgradeTask):
    description = "Check Python source files syntax sugars with pyupgrade."
    python_dependencies = ["pyupgrade"]

    keep_runtime_typing: Property[bool] = Property.default(False)
    additional_files: Property[Sequence[Path]] = Property.default_factory(list)
    python_version: Property[str]

    def execute(self) -> TaskStatus:
        # We copy the file because there is no way to make pyupgrade not edit the files
        old_dir = self.settings.project.directory.resolve()
        new_file_for_old_file = {}
        with TemporaryDirectory() as new_dir:
            for file in self.additional_files.get():
                new_file = new_dir / file.resolve().relative_to(old_dir)
                new_file.parent.mkdir(parents=True, exist_ok=True)
                new_file.write_bytes(file.read_bytes())
                new_file_for_old_file[file] = new_file
            self._files = new_file_for_old_file.values()

            result = super().execute()
            if not result.is_failed():
                return result  # nothing more to do

            # We print a diff
            for old_file, new_file in new_file_for_old_file.items():
                old_content = old_file.read_text()
                new_content = new_file.read_text()
                if old_content != new_content:
                    stdout.writelines(
                        unified_diff(
                            old_content.splitlines(keepends=True),
                            new_content.splitlines(keepends=True),
                            fromfile=str(old_file),
                            tofile=str(old_file),
                            n=5,
                        )
                    )
            return result

    def get_execute_command(self) -> list[str]:
        return self.run_pyupgrade(self._files, ())


@dataclasses.dataclass
class PyUpgradeTasks:
    check: PyUpgradeTask
    format: PyUpgradeTask


def pyupgrade(
    *,
    name: str = "python.pyupgrade",
    project: Project | None = None,
    exclude: Collection[Path] = (),
    exclude_patterns: Collection[str] = (),
    keep_runtime_typing: bool = False,
    python_version: str = "3",
    additional_files: Sequence[Path] = (),
) -> PyUpgradeTasks:
    project = project or Project.current()
    settings = python_settings(project)

    directories = [
        p for p in (*additional_files, settings.source_directory, settings.get_tests_directory()) if p is not None
    ]
    files = {f.resolve() for p in directories for f in Path(p).glob("**/*.py")}
    exclude = [e.resolve() for e in exclude]
    filtered_files = [
        f
        for f in files
        if not any(_is_relative_to(f, i) for i in exclude) and not any(f.match(p) for p in exclude_patterns)
    ]

    check_task = project.task(f"{name}.check", PyUpgradeCheckTask, group="lint")
    check_task.additional_files = filtered_files
    check_task.keep_runtime_typing = keep_runtime_typing
    check_task.python_version = python_version

    format_task = project.task(name, PyUpgradeTask, group="fmt")
    format_task.additional_files = filtered_files
    format_task.keep_runtime_typing = keep_runtime_typing
    format_task.python_version = python_version

    return PyUpgradeTasks(check_task, format_task)


def _is_relative_to(a: Path, b: Path) -> bool:
    # Polyfill for Python 3.7 and 3.8
    try:
        a.relative_to(b)
        return True
    except ValueError:
        return False



""" Abstraction of Python build systems such as Poetry and Slap. """


from __future__ import annotations

import abc
import logging
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar

from kraken.core import TaskStatus

from kraken.std.python.pyproject import Pyproject, PyprojectHandler

if TYPE_CHECKING:
    from ..settings import PythonSettings

logger = logging.getLogger(__name__)


class PythonBuildSystem(abc.ABC):
    """Abstraction of a Python build system."""

    name: ClassVar[str]

    @abc.abstractmethod
    def supports_managed_environments(self) -> bool:
        """Return `True` if the build system supports managed environments."""

    @abc.abstractmethod
    def get_managed_environment(self) -> ManagedEnvironment:
        """Return a handle for the managed environment.

        :raise NotImplementedError: If :meth:`supports_managed_environment` returns `False`.
        """

    def update_pyproject(self, settings: PythonSettings, pyproject: Pyproject) -> None:
        """A chance to permanently update the Pyproject configuration."""

        handler = self.get_pyproject_reader(pyproject)
        package_sources = [x for x in settings.package_indexes.values() if x.is_package_source]
        if package_sources:
            try:
                handler.set_package_indexes(package_sources)
            except NotImplementedError:
                logger.debug("build system %r does not support managing package indexes", self.name)

    @abc.abstractmethod
    def update_lockfile(self, settings: PythonSettings, pyproject: Pyproject) -> TaskStatus:
        """Resolve all dependencies of the project and write the exact versions into
        the correspondig lock file. In the case of Poetry it is poetry.lock."""

    @abc.abstractmethod
    def requires_login(self) -> bool:
        """Return True if this build system requires a separate login step (i.e. if the credentials cannot be
        passed at install time to the command-line)."""

    def login(self, settings: PythonSettings) -> None:
        """Log into the Python package indices."""

        raise NotImplementedError

    @abc.abstractmethod
    def build(self, output_directory: Path, as_version: str | None = None) -> list[Path]:
        """Build one or more distributions of the project managed by this build system.

        :param output_directory: The directory where the distributions should be placed.
        :param as_version: A version number for the built distributions.
        """

    @abc.abstractmethod
    def get_pyproject_reader(self, pyproject: Pyproject) -> PyprojectHandler:
        """Return an object able to read the pyproject file depending on the build system."""


class ManagedEnvironment(abc.ABC):
    """Abstraction of a managed Python environment."""

    @abc.abstractmethod
    def exists(self) -> bool:
        """Return `True` if the managed environment exists."""

    @abc.abstractmethod
    def get_path(self) -> Path:
        """Return the path to the managed environment.

        :raises RuntimeError: May be raised if the environment does not exist (it may not be possible to
            determine the path of the environment before it exists depending on the build system).
        """

    @abc.abstractmethod
    def install(self, settings: PythonSettings) -> None:
        """Install the managed environment. This should be a no-op if the environment already exists."""

    def always_install(self) -> bool:
        """Always trigger environment installation even if already exists"""
        return False


def detect_build_system(project_directory: Path) -> PythonBuildSystem | None:
    """Detect the Python build system used in *project_directory*."""

    pyproject_toml = project_directory / "pyproject.toml"
    if not pyproject_toml.is_file():
        return None

    pyproject_content = pyproject_toml.read_text()

    if "[tool.slap]" in pyproject_content:
        from .slap import SlapPythonBuildSystem

        return SlapPythonBuildSystem(project_directory)

    if "poetry-core" in pyproject_content:
        from .poetry import PoetryPythonBuildSystem

        return PoetryPythonBuildSystem(project_directory)

    if "maturin" in pyproject_content:
        if "[tool.poetry]" in pyproject_content:
            from .maturin import MaturinPoetryPythonBuildSystem

            return MaturinPoetryPythonBuildSystem(project_directory)
        else:
            from .maturin import MaturinPdmPythonBuildSystem

            return MaturinPdmPythonBuildSystem(project_directory)

    if "pdm" in pyproject_content:
        from .pdm import PDMPythonBuildSystem

        return PDMPythonBuildSystem(project_directory)

    return None



import errno
import json
import logging
import os
import signal
import subprocess
import sys
import time
from collections.abc import Callable, Mapping
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import IO, Any, Literal

logger = logging.getLogger(__name__)


def process_exists(pid: int) -> bool:
    """Checks if the processed with the given *pid* exists. Returns #True if
    that is the case, #False otherwise."""

    if pid == 0:
        return False
    try:
        os.kill(pid, 0)
    except OSError as exc:
        if exc.errno == errno.ESRCH:
            return False
    return True


def wait_for_child_process(pid: int, timeout: float, interval: float = 0.1) -> int | None:
    """Wait for the process of *pid* to exit and return the exit code. This only works for child processes."""

    tstart = time.perf_counter()
    while (time.perf_counter() - tstart) < timeout:
        (pid_returned, status) = os.waitpid(pid, os.WNOHANG)
        if pid_returned != 0:
            return os.waitstatus_to_exitcode(status)
        time.sleep(interval)
    return None


def wait_for_process(pid: int, timeout: float, interval: float = 0.1) -> bool:
    """Wait for a process to exit. Returns True if the process exited within the timeout period."""

    tstart = time.perf_counter()
    while (time.perf_counter() - tstart) < timeout:
        if not process_exists(pid):
            return True
        time.sleep(interval)
    return False


def process_terminate(
    pid: int,
    allow_kill: bool = True,
    sigint_timeout: float = 3.0,
    sigterm_timeout: float = 2.0,
    sigkill_timeout: float = 1.0,
) -> bool:
    """Terminates the process with the given *pid*. First sends #signal.SIGINT,
    followed by #signal.SIGTERM after *sigint_timeout* seconds, followed by
    #signal.SIGKILL after *sigkill_timeout* seconds if the process has not responded to
    the terminate signal.

    The fallback to kill can be disabled by setting *allow_kill* to False.
    Returns True if the process was successfully terminated or killed, or if
    the process did not exist in the first place."""

    try:
        os.kill(pid, signal.SIGINT)
        if wait_for_process(pid, sigint_timeout) is not None:
            return True
        os.kill(pid, signal.SIGTERM)
        if wait_for_process(pid, sigterm_timeout) is not None:
            return True
        if allow_kill:
            os.kill(pid, signal.SIGKILL)
            return wait_for_process(pid, sigkill_timeout) is not None
        return False
    except OSError as exc:
        if exc.errno == errno.ESRCH:
            return True
        raise


def spawn_fork(func: Callable[[], Any], detach: bool = True) -> int:
    """Spawns a single fork process and calls *func*. If *detach* is #True,
    the fork will be detached first (note that this process will still be killed
    by it's parent process if it doesn't exist gracefully).

    This is useful if *func* spawns another process, which will then behave like
    a daemon (as it will NOT be killed if the original process dies)."""

    if not callable(func):
        raise TypeError(f"func is of type {type(func).__name__} which is not callable")

    pid = os.fork()
    if pid > 0:
        # Return to the original caller
        return pid
    if detach:
        os.setsid()
    func()
    os._exit(os.EX_OK)


def replace_stdio(stdin: int | None = None, stdout: int | None = None, stderr: int | None = None) -> None:
    """Replaces the file handles of stdin/sdout/stderr, closing the original file descriptors if necessary."""

    if stdin is not None:
        os.dup2(stdin, sys.stdin.fileno())
    if stdout is not None:
        os.dup2(stdout, sys.stdout.fileno())
    if stderr is not None:
        os.dup2(stderr, sys.stderr.fileno())


class DaemonController:
    """
    Manages a daemon process. If the process isn't running, or the configuration of the process changed, it will be
    restarted.

    :param state_file: The file where the daemon state, such as the PID and the arguments that the daemon was started
        with, is stored. If the file does not exist, it will be assumed that the daemon is not running. This file
        is mandatory for the controller.
    """

    @dataclass(frozen=True)
    class State:
        command: list[str]
        cwd: str
        env: dict[str, str]
        pid: int
        started_at: float

    def __init__(self, name: str, state_file: Path) -> None:
        self.name = name
        self.state_file = state_file

    def load_state(self) -> State | None:
        try:
            data = json.loads(self.state_file.read_text())
        except FileNotFoundError:
            return None
        except json.JSONDecodeError:
            logger.warning("Could not decode Daemon %r state file. Continuing as if it did not exist.", self.name)
            return None
        try:
            return self.State(**data)
        except TypeError:
            if "pid" in data and isinstance(data["pid"], int):
                logger.warning(
                    "Could not initialize Daemon %r state from payload, but a `pid` field was found. Continuing "
                    "with a state that does not contain any information besides the `pid`.",
                    self.name,
                )
                return self.State([], os.getcwd(), {}, data["pid"], 0)
            else:
                logger.error(
                    "Could not recover from invalid JSON payload found for Daemon %r state. Continuing as if it "
                    "did not exist.",
                    self.name,
                )
            return None

    def save_state(self, state: State) -> None:
        self.state_file.write_text(json.dumps(asdict(state)))

    def remove_state(self) -> None:
        try:
            self.state_file.unlink()
        except FileNotFoundError:
            pass

    def is_alive(self) -> bool:
        state = self.load_state()
        return bool(state is not None and process_exists(state.pid))

    def run(
        self,
        command: list[str],
        cwd: Path | None = None,
        env: Mapping[str, str] | None = None,
        stdout: Path | None = None,
        stderr: Path | None | Literal["stdout"] = "stdout",
        mkdir: bool = True,
    ) -> bool:
        """
        Ensure that the daemon is running. If the command or cwd changed, the daemon will be restarted. Returns True
        if the daemon was started, False if it was already running.
        """

        if stderr is not None and stderr != "stdout":
            assert isinstance(stderr, Path), type(stderr)
            stderr_path = stderr.absolute()
        else:
            stderr_path = None
        stderr_to_stdout = stderr == "stdout"

        cwd = cwd.absolute() if cwd else Path.cwd()
        env = dict(env.items()) if env else {}
        stdout = stdout.absolute() if stdout else None

        state = self.load_state()
        daemon_alive = bool(state and process_exists(state.pid))
        if daemon_alive and state and state.command == command and state.cwd == str(cwd) and env == state.env:
            logger.info("Daemon %r is already running (pid: %s)", self.name, state.pid)
            return False

        if state and daemon_alive:
            logger.info("Restarting daemon %r.", self.name)
            process_terminate(state.pid)
            self.remove_state()
            state = None
        else:
            logger.info("Starting daemon %r.", self.name)

        def start_daemon() -> None:
            if stdout is not None:
                if mkdir:
                    stdout.parent.mkdir(parents=True, exist_ok=True)
                stdout_fp: IO[Any] | int = stdout.open("wb")
            else:
                stdout_fp = subprocess.DEVNULL

            if stderr_to_stdout:
                stderr_fp: IO[Any] | int = stdout_fp
            elif stderr_path:
                if mkdir:
                    stderr_path.parent.mkdir(parents=True, exist_ok=True)
                stderr_fp = stderr_path.open("wb")
            else:
                stderr_fp = subprocess.DEVNULL

            proc_env = os.environ.copy()
            proc_env.update(env)
            proc = subprocess.Popen(command, cwd=cwd, env=proc_env, stdout=stdout_fp, stderr=stderr_fp)
            logger.info("Daemon %r started (pid: %s)", self.name, proc.pid)
            state = self.State(command, str(cwd), env, proc.pid, time.time())
            self.save_state(state)

        pid = spawn_fork(start_daemon)
        if status := wait_for_child_process(pid, 10.0):
            raise Exception("An unexpected error ocurred when starting daemon %r (exit code %s).", self.name, status)
        return True

    def stop(self) -> bool:
        """
        Stop the daemon if it is currently running. Returns True if the daemon was running and is now stopped.
        """

        state = self.load_state()
        if state and process_exists(state.pid):
            logger.info("Stopping daemon %r.", self.name)
            process_terminate(state.pid)
            self.remove_state()
            return True

        return False


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="[%(asctime)s] %(message)s")
    controller = DaemonController("cli", Path("cli-daemon.state"))
    if len(sys.argv) == 1:
        if not controller.stop():
            logger.warning("Daemon was not running.")
    else:
        controller.run(sys.argv[1:], Path.cwd(), {}, Path("cli-daemon.txt"), "stdout")



from kraken.std.python.buildsystem.pdm import PdmPyprojectHandler
from kraken.std.python.pyproject import PackageIndex, Pyproject

EXAMPLE_PDM_PYPROJECT = """
[project]
name = "pdm-project"
version = "0.1.0"

[[tool.pdm.source]]
name = "private"
url = "https://private.pypi.org/simple"
verify_ssl = true
"""


def test__PdmPyprojectHandler__getters() -> None:
    handler = PdmPyprojectHandler(Pyproject.read_string(EXAMPLE_PDM_PYPROJECT))
    assert handler.get_name() == "pdm-project"
    assert handler.get_version() == "0.1.0"


def test__PdmPyprojectHandler__set_version() -> None:
    handler = PdmPyprojectHandler(Pyproject.read_string(EXAMPLE_PDM_PYPROJECT))
    handler.set_version("2.0.0")
    assert handler.raw["project"]["version"] == "2.0.0"


def test__PdmPyprojectHandler__get_package_indexes() -> None:
    handler = PdmPyprojectHandler(Pyproject.read_string(EXAMPLE_PDM_PYPROJECT))
    assert handler.get_package_indexes() == [
        PackageIndex(
            alias="private",
            index_url="https://private.pypi.org/simple",
            priority=PackageIndex.Priority.primary,
            verify_ssl=True,
        )
    ]


def test__PdmPyprojectHandler__set_package_indexes__to_empty_list() -> None:
    handler = PdmPyprojectHandler(Pyproject.read_string(EXAMPLE_PDM_PYPROJECT))
    handler.set_package_indexes([])
    assert handler.raw["tool"]["pdm"]["source"] == []


def test__PdmPyprojectHandler__set_package_indexes__to_various_indexes() -> None:
    handler = PdmPyprojectHandler(Pyproject.read_string(EXAMPLE_PDM_PYPROJECT))
    handler.set_package_indexes(
        [
            PackageIndex("a", "https://a.com", PackageIndex.Priority.primary, True),
            PackageIndex("c", "https://c.com", PackageIndex.Priority.supplemental, False),
            PackageIndex("b", "https://b.com", PackageIndex.Priority.default, True),
        ]
    )
    assert handler.raw["tool"]["pdm"]["source"] == [
        {"name": "b", "url": "https://b.com"},
        {"name": "a", "url": "https://a.com"},
        {"name": "c", "url": "https://c.com", "verify_ssl": False},
    ]



import subprocess
from dataclasses import dataclass
from pathlib import Path

from kraken.core import Property, Task, TaskStatus

from kraken.std.descriptors.resource import LibraryArtifact


@dataclass
class CargoDebianArtifact(LibraryArtifact):
    pass


class CargoGenerateDebPackage(Task):
    package_name: Property[str]
    out_packages: Property[list[CargoDebianArtifact]] = Property.output()

    # Task

    def prepare(self) -> TaskStatus:
        return TaskStatus.pending()

    def execute(self) -> TaskStatus:
        package_name = self.package_name.get()
        command = ["cargo", "install", "cargo-deb", "--force"]
        result = subprocess.call(command)
        if result != 0:
            return TaskStatus.from_exit_code(command, result)

        output_deb = f"{package_name}.deb"
        command = ["cargo", "deb", "--output", output_deb]
        result = subprocess.call(command)
        if result != 0:
            return TaskStatus.from_exit_code(command, result)

        out_packages: list[CargoDebianArtifact] = [CargoDebianArtifact(package_name, Path(output_deb))]
        self.out_packages.set(out_packages)
        return TaskStatus.succeeded()



from collections.abc import Sequence

from .parser import GitignoreFile

GENERATED_SECTION_START = "# GENERATED-CONTENT-START"
GENERATED_SECTION_END = "# GENERATED-CONTENT-END"


def split_generated_section(file: GitignoreFile) -> tuple[GitignoreFile, GitignoreFile, GitignoreFile]:
    """
    Splits a GitignoreFile into three parts: the generated section surrounded by the user sections.

    The generated section is marked to start and end with the following comments:

    ```
    # GENERATED-CONTENT-START
    ...
    # GENERATED-CONTENT-END
    ```

    In addition, we support the following format for backwards compatibility for `.gitignore` files
    managed by Kraken 0.26.1 and before:

    ```
    ### START-GENERATED-CONTENT [HASH: ...]

    ### END-GENERATED-CONTENT
    ```

    The middle `GitignoreFile` that is returned will not contain these start and end markers.
    """

    user1 = GitignoreFile()
    generated = GitignoreFile()
    user2 = GitignoreFile()
    target = user1

    it = iter(file)
    entry = next(it, None)
    while entry is not None:
        if entry.type == "comment" and entry.strip() == GENERATED_SECTION_START:
            if target is not user1:
                raise ValueError("Multiple GENERATED-CONTENT-START markers found")
            target = generated
        elif entry.type == "comment" and entry.strip() == GENERATED_SECTION_END:
            if target is not generated:
                raise ValueError("GENERATED-CONTENT-END marker found out of order")
            target = user2

        # Also handle the old format:
        elif entry.type == "comment" and entry.startswith("### START-GENERATED-CONTENT"):
            if target is not user1:
                raise ValueError("Multiple legacy START-GENERATED-CONTENT markers found")
            target = generated
            while True:
                entry = next(it, None)
                if not entry or entry.type != "comment" or entry.startswith("### [PARAMETERS_HASH:"):
                    break
        elif entry.type == "comment" and entry.strip() == "### END-GENERATED-CONTENT":
            if target is not generated:
                raise ValueError("Legacy END-GENERATED-CONTENT marker found out of order")
            if generated and generated[-1].type == "comment" and generated[-1].startswith("# ----"):
                generated.pop()
            target = user2

        else:
            target.append(entry)

        entry = next(it, None)

    return user1, generated, user2


def join_generated_section(user1: Sequence[str], generated: Sequence[str], user2: Sequence[str]) -> GitignoreFile:
    return GitignoreFile.parse([*user1, GENERATED_SECTION_START, *generated, GENERATED_SECTION_END, *user2])



"""Create distributions (archives) from build artifacts and resources."""

from __future__ import annotations

import abc
import logging
import tarfile
import zipfile
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from fnmatch import fnmatch
from pathlib import Path
from typing import Any, Literal, Union, cast

import databind.json
from kraken.core import Project, Property, Task, TaskSet
from termcolor import colored

from .descriptors.resource import BinaryArtifact, LibraryArtifact, Resource

logger = logging.getLogger(__name__)


@dataclass
class IndividualDistOptions:
    arcname: str | None = None
    exclude: Sequence[str] = ()
    include: Sequence[str] | None = None


@dataclass
class ConfiguredResource(Resource):
    options: IndividualDistOptions


class DistributionTask(Task):
    """Create an archive from a set of files and resources."""

    #: The output filename. If a string is specified, that string is treated relative
    #: to the project output directory. A Path object is treated relative to the project
    #: directory. Unless the :attr:`archive_type` property is set, the suffix will determine
    #: the type of archive that is created.
    output_file: Property[Path]

    #: The type of archive that will be created. Can be zip, tgz and tar.
    archive_type: Property[str]

    #: Prefix to add to all files added to the distribution.
    prefix: Property[str]

    #: A list of resources to include.
    resources: Property[list[ConfiguredResource]] = Property.default_factory(list)

    #: A resource that describes the output file.
    _output_file_resource: Property[Resource] = Property.output()

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self._output_file_resource.set(self.output_file.map(lambda p: Resource("dist", p)))

    # Task

    def execute(self) -> None:
        output_file = self.output_file.get()
        archive_type = self.archive_type.get_or(None)
        if archive_type is None:
            archive_type = output_file.suffix.lstrip(".")
            archive_type = {"tgz": "tar.gz", "txz": "tar.xz", "tbz2": "tar.bz2"}.get(archive_type, archive_type)
        assert isinstance(archive_type, str)

        print("Writing archive", colored(str(output_file), "yellow"))
        output_file.parent.mkdir(parents=True, exist_ok=True)
        with wopen_archive(output_file, archive_type) as archive:
            for resource in self.resources.get():
                arcname = resource.options.arcname
                if arcname is None:
                    arcname = str(resource.path.relative_to(self.project.directory))
                print(
                    "  +",
                    colored(arcname or ".", "green"),
                    f"({resource.path})" if arcname != str(resource.path) else "",
                )
                add_to_archive(
                    archive,
                    arcname,
                    self.project.directory / resource.path,
                    resource.path,
                    resource.options.exclude,
                    resource.options.include,
                )


def wopen_archive(path: Path, type_: str) -> ArchiveWriter:
    """Open an archive at *path* for writing. The *type_* indicates what type of archive will be created.
    Accepted values for *type_* are `zip`, `tar`, `tar.gz`, `tar.bz2` and `tar.xz`."""

    if type_.startswith("tar."):
        return TarArchiveWriter(path, cast(Any, type_.partition(".")[-1]))
    elif type_ == "tar":
        return TarArchiveWriter(path, "")
    elif type_ == "zip":
        return ZipArchiveWriter(path)
    else:
        raise ValueError(f"unsupported archive type: {type_!r}")


def add_to_archive(
    writer: ArchiveWriter,
    arcname: str,
    path: Path,
    test_path: Path | None = None,
    exclude: Sequence[str] = (),
    include: Sequence[str] | None = None,
) -> None:
    """Recursively adds *path* to the archive *writer* under consideration of the *exclude* and *include* glob
    patterns that are tested against *test_path*.

    The glob patterns for the *exclude* and *include* arguments are tested against the full relative *test_path*
    as well as the individual filename. (NOTE(niklas.rosenstein): To support full .gitignore like behaviour, we'd
    need to test the entire range from name to full path).

    :param writer: The archive writer implementation.
    :param arcname: The name to give *path*. Sub-paths are appended to this name as normal.
    :param path: The path to write to the archive.
    :param test_path: The path to test the exclude/include patterns against. Sub-paths are appended to this as normal.
        Defaults to *path*.
    :param exclude: A sequence of glob patterns that, if matching, cause an element (file or directory) to be excluded.
    :param include: If specified, must be a sequence of glob patterns that will cause a file or directory to only be
        added if any pattern matches, and no *exclude* pattern matches.
    """

    test_path = test_path or path
    s_test_path = str(test_path)

    if any(fnmatch(s_test_path, x) or fnmatch(test_path.name, x) for x in exclude):
        return
    if include is not None and not any(fnmatch(s_test_path, x) or fnmatch(test_path.name, x) for x in include):
        return

    if path.is_dir():
        for item in path.iterdir():
            add_to_archive(writer, arcname + "/" + item.name, item, test_path / item.name, exclude, include)
    else:
        writer.add_file(arcname, path)


class ArchiveWriter(abc.ABC):
    """Base class to write an archive file."""

    @abc.abstractmethod
    def add_file(self, arcname: str, path: Path) -> None:
        """Add a file to the archive."""

    @abc.abstractmethod
    def close(self) -> None:
        pass

    def add_path(self, arcname: str, path: Path) -> None:
        """Recursively add the contents of all files under *path*."""

        if path.is_dir():
            for item in path.iterdir():
                self.add_path(arcname + "/" + item.name, item)
        else:
            self.add_file(arcname, path)

    def __enter__(self) -> ArchiveWriter:
        return self

    def __exit__(self, *a: Any) -> None:
        self.close()


class TarArchiveWriter(ArchiveWriter):
    def __init__(self, path: Path, type_: Literal["", "gz", "bz2", "xz"]) -> None:
        self._archive = tarfile.open(path, mode="w:" + type_)

    def close(self) -> None:
        self._archive.close()

    def add_file(self, arcname: str, path: Path) -> None:
        self._archive.add(path, arcname, recursive=False)


class ZipArchiveWriter(ArchiveWriter):
    def __init__(self, path: Path) -> None:
        self._archive = zipfile.ZipFile(path, "w")

    def close(self) -> None:
        self._archive.close()

    def add_file(self, arcname: str, path: Path) -> None:
        self._archive.write(path, arcname)


def dist(
    *,
    name: str,
    dependencies: Sequence[str | Task] | Mapping[str, Mapping[str, Any] | IndividualDistOptions],
    output_file: str | Path,
    archive_type: str | None = None,
    prefix: str | None = None,
    project: Project | None = None,
) -> DistributionTask:
    """Create a task that produces a distribution from the resources provided by the tasks specified with *include*.

    :param name: The name of the task.
    :param dependencies: A list of tasks or task selectors (resolved relative to the *project*) that provide the
        resources that should be included in the generated archive. If a dictionary is specified, instead, it must
        map each dependency to a dictionary of settings that can be deserialized into :class:`IndividualDistOptions`.
    :param output_file: The output filename to write to. If a string is specified, it will be treated relative to the
        project build directory. If a path object is specified, it will be treated relative to the project directory.
    :param archive_type: The type of archive to create (e.g. `zip`, `tar`, `tar.gz`, `tar.bz2` or `tar.xz`). If left
        empty, the archive type is derived from the *output_filename* suffix.
    :param project: The project to create the task for.
    """

    project = project or Project.current()

    if isinstance(output_file, str):
        output_file = project.build_directory / output_file
    else:
        output_file = project.directory / output_file

    if isinstance(dependencies, Sequence):
        dependencies = cast(
            Mapping[str, Union[Mapping[str, Any], IndividualDistOptions]], {d: {} for d in dependencies}
        )
    dependencies_map = {
        k: databind.json.load(v, IndividualDistOptions) if not isinstance(v, IndividualDistOptions) else v
        for k, v in dependencies.items()
    }
    dependencies_set = project.resolve_tasks(dependencies_map)

    # This associates the IndividualDistOptions specified in *dependencies* to the Resource(s)
    # provided by the task(s).
    resources = (
        dependencies_set.select(Resource)
        .dict_supplier()
        .map(
            lambda resources: get_configured_resources(
                resources,
                dependencies_map,
                dependencies_set,
            )
        )
    )

    dist_task = project.task(name, DistributionTask)
    dist_task.resources = resources
    dist_task.output_file = output_file
    dist_task.archive_type = archive_type
    dist_task.prefix = prefix
    return dist_task


def get_configured_resources(
    resources: dict[Task, list[Resource]],
    dependencies_map: dict[str, IndividualDistOptions],
    dependencies_set: TaskSet,
) -> list[ConfiguredResource]:
    configured_resources = []

    for task, task_resources in resources.items():
        for resource in task_resources:
            task_options = dependencies_map[next(iter(dependencies_set.partitions()[task]))]
            options = IndividualDistOptions(**vars(task_options))

            if options.arcname is None and isinstance(resource, (BinaryArtifact, LibraryArtifact)):
                options.arcname = resource.name
            configured_resources.append(ConfiguredResource(**vars(resource), options=options))
    return configured_resources



""" Implements Maturin as a build system for kraken-std. """

from __future__ import annotations

import logging
import os
import shutil
import subprocess as sp
from collections.abc import Callable, Collection
from dataclasses import dataclass
from pathlib import Path

from kraken.common.path import is_relative_to

from ...cargo.manifest import CargoMetadata
from ..pyproject import Pyproject, PyprojectHandler
from ..settings import PythonSettings
from . import ManagedEnvironment
from .pdm import PDMManagedEnvironment, PDMPythonBuildSystem
from .poetry import PoetryManagedEnvironment, PoetryPyprojectHandler, PoetryPythonBuildSystem

logger = logging.getLogger(__name__)


@dataclass
class MaturinZigTarget:
    """A specific target to build for with Maturin.
    :param target: Rust target to cross-compile to using zig.
        For example "x86_64-unknown-linux-gnu" or "aarch64-apple-darwin".
        Requires the `maturin[zig]` pip package.
        These targets should be installed into the Rust installation.
    :param zig_features: Cargo features to enable for zig builds. If zig is used, it should be at least `pyo3/abi3`
        or another feature depending on `pyo3/abi3` (`pyo3/abi3-py38`...).
    :param manylinux: if set to true, this will produce a manylinux wheel, and any dynamically
        linked libraries will be copied into the wheel. If false, wheels will be tagged as
        'linux' and dynamically linked libraries are the responsibility of the user.
    :param macos_sdk_root: For zig builds targeting macOS, the path to the MacOS SDK to use.
        By default, the `SDKROOT` environment variable is used as a fallback.
        Only required when cross compiling from Linux to MacOS.
    :param rustflags: RUSTFLAGS environment variable will be set at compilation time. This can be
        used to add e.g. native libraries to link against.
    :param ld_library_path: LD_LIBRARY_PATH environment variable will be set at compilation time. This can be
        used to add any native libraries that might be required by pypa fixups so to produce manylinux wheels.
        Likely the same content as RUSTFLAGS, but in the LD_LIBRARY_PATH format.
    """

    target: str
    rustflags: str | None = None
    ld_library_path: str | None = None
    macos_sdk_root: Path | None = None
    manylinux: bool = True
    zig_features: Collection[str] = ()


class _MaturinBuilder:
    def __init__(
        self, entry_point: str, get_pyproject_reader: Callable[[Pyproject], PyprojectHandler], project_directory: Path
    ) -> None:
        self._entry_point = entry_point
        self._get_pyproject_reader = get_pyproject_reader
        self._project_directory = project_directory
        self._default_build = True
        self._zig_targets: Collection[MaturinZigTarget] = []

    def disable_default_build(self) -> None:
        self._default_build = False

    def enable_zig_build(self, targets: Collection[MaturinZigTarget]) -> None:
        """
        :param targets: Collection of MaturinTargets to cross-compile to using zig.
        """
        self._zig_targets = targets

    def build(self, output_directory: Path, as_version: str | None) -> list[Path]:
        # We set the version
        old_poetry_version = None
        pyproject_path = self._project_directory / "pyproject.toml"
        if as_version is not None:
            pyproject = self._get_pyproject_reader(Pyproject.read(pyproject_path))
            old_poetry_version = pyproject.get_version()
            pyproject.set_version(as_version)
            pyproject.raw.save()

        # We clean up target dir
        metadata = CargoMetadata.read(self._project_directory)
        dist_dir = metadata.target_directory / "wheels"
        if dist_dir.exists():
            shutil.rmtree(dist_dir)

        # We run the actual build
        if self._default_build:
            command = [self._entry_point, "run", "maturin", "build", "--release"]
            logger.info("%s", command)
            sp.check_call(command, cwd=self._project_directory)
        for target in self._zig_targets:
            command = [
                self._entry_point,
                "run",
                "maturin",
                "build",
                "--release",
                "--zig",
                "--target",
                target.target,
                "--features",
                ",".join(target.zig_features),
            ]
            if not target.manylinux:
                command.append("--manylinux")
                command.append("off")
            logger.info("%s", command)
            env = os.environ.copy()
            if target.target.endswith("-apple-darwin"):
                if target.macos_sdk_root is not None:
                    env["SDKROOT"] = str(target.macos_sdk_root.resolve())
                elif "SDKROOT" not in env:
                    logger.error(f"No macOS SDKROOT set for the target {target}")
            if target.rustflags is not None:
                env["RUSTFLAGS"] = target.rustflags
            if target.ld_library_path is not None:
                env["LD_LIBRARY_PATH"] = target.ld_library_path
            sp.check_call(command, cwd=self._project_directory, env=env)

        # We get the output files
        src_files = list(dist_dir.iterdir())
        dst_files = [output_directory / path.name for path in src_files]
        for src, dst in zip(src_files, dst_files):
            shutil.move(str(src), dst)

        # Unless the output directory is a subdirectory of the dist_dir, we remove the dist dir again.
        if not is_relative_to(output_directory, dist_dir):
            shutil.rmtree(dist_dir)

        if as_version is not None:
            # We roll back the version
            pyproject = self._get_pyproject_reader(Pyproject.read(pyproject_path))
            pyproject.set_version(old_poetry_version)
            pyproject.raw.save()

        return dst_files


class MaturinPoetryPyprojectHandler(PoetryPyprojectHandler):
    def set_version(self, version: str | None) -> None:
        PyprojectHandler.set_version(self, version)
        PoetryPyprojectHandler.set_version(self, version)

    def synchronize_project_section_to_poetry_state(self) -> None:
        """
        Synchronize the `[tool.poetry]` package metadata to the `[project]` section.

        In the case where only the `[project]` section exists, we sync it into the other direction.
        """

        poetry_section = self._poetry_section
        project_section = self.raw.setdefault("project", {})
        for field_name in ("name", "version"):
            poetry_value = poetry_section.get(field_name)
            project_value = project_section.get(field_name)
            if poetry_value is None:
                poetry_section[field_name] = project_value
            else:
                project_section[field_name] = poetry_value


class MaturinPoetryPythonBuildSystem(PoetryPythonBuildSystem):
    """A maturin-backed version of the Poetry build system, that invokes the maturin build-backend.
    Can be enabled by adding the following to the local pyproject.yaml:
    ```toml
    [tool.poetry.dev-dependencies]
    maturin = "^1.0"

    [build-system]
    requires = ["maturin~=1.0"]
    build-backend = "maturin"
    ```
    """

    name = "Maturin Poetry"

    def __init__(self, project_directory: Path) -> None:
        super().__init__(project_directory)
        self._builder = _MaturinBuilder("poetry", self.get_pyproject_reader, self.project_directory)

    def disable_default_build(self) -> None:
        self._builder.disable_default_build()

    def enable_zig_build(self, targets: Collection[MaturinZigTarget]) -> None:
        """
        :param targets: Collection of MaturinTargets to cross-compile to using zig.
        """
        self._builder.enable_zig_build(targets)

    # PythonBuildSystem

    def get_pyproject_reader(self, pyproject: Pyproject) -> MaturinPoetryPyprojectHandler:
        return MaturinPoetryPyprojectHandler(pyproject)

    def get_managed_environment(self) -> ManagedEnvironment:
        return MaturinPoetryManagedEnvironment(self.project_directory)

    def update_pyproject(self, settings: PythonSettings, pyproject: Pyproject) -> None:
        super().update_pyproject(settings, pyproject)
        handler = self.get_pyproject_reader(pyproject)
        handler.synchronize_project_section_to_poetry_state()

    def build(self, output_directory: Path, as_version: str | None = None) -> list[Path]:
        return self._builder.build(output_directory, as_version)


class MaturinPoetryManagedEnvironment(PoetryManagedEnvironment):
    def install(self, settings: PythonSettings) -> None:
        super().install(settings)
        command = ["poetry", "run", "maturin", "develop"]
        logger.info("%s", command)
        sp.check_call(command, cwd=self.project_directory)

    def always_install(self) -> bool:
        return True


class MaturinPdmPythonBuildSystem(PDMPythonBuildSystem):
    """A maturin-backed version of the PDM build system, that invokes the maturin build-backend.
    Can be enabled by adding the following to the local pyproject.yaml:
    ```toml
    [tool.pdm.dev-dependencies]
    build = ["maturin~=1.0", "pip~=23.0"]

    [build-system]
    requires = ["maturin~=1.0"]
    build-backend = "maturin"
    ```
    """

    name = "Maturin PDM"

    def __init__(self, project_directory: Path) -> None:
        super().__init__(project_directory)
        self._builder = _MaturinBuilder("pdm", self.get_pyproject_reader, self.project_directory)

    def disable_default_build(self) -> None:
        self._builder.disable_default_build()

    def enable_zig_build(self, targets: Collection[MaturinZigTarget]) -> None:
        """
        :param targets: Collection of MaturinTargets to cross-compile to using zig.
        """
        self._builder.enable_zig_build(targets)

    def get_managed_environment(self) -> ManagedEnvironment:
        return MaturinPdmManagedEnvironment(self.project_directory)

    def build(self, output_directory: Path, as_version: str | None = None) -> list[Path]:
        return self._builder.build(output_directory, as_version)


class MaturinPdmManagedEnvironment(PDMManagedEnvironment):
    def always_install(self) -> bool:
        return True



from .settings import PythonSettings, python_settings
from .tasks.black_task import BlackTask, black
from .tasks.build_task import BuildTask, build
from .tasks.flake8_task import Flake8Task, flake8
from .tasks.info_task import InfoTask, info
from .tasks.install_task import InstallTask, install
from .tasks.isort_task import IsortTask, isort
from .tasks.login_task import login
from .tasks.mypy_stubtest_task import MypyStubtestTask, mypy_stubtest
from .tasks.mypy_task import MypyTask, mypy
from .tasks.publish_task import PublishTask, publish
from .tasks.pycln_task import PyclnTask, pycln
from .tasks.pylint_task import PylintTask, pylint
from .tasks.pytest_task import CoverageFormat, PytestTask, pytest
from .tasks.pyupgrade_task import PyUpgradeCheckTask, PyUpgradeTask, pyupgrade
from .tasks.update_lockfile_task import update_lockfile_task
from .tasks.update_pyproject_task import update_pyproject_task
from .version import git_version_to_python_version

# Backwards compatibilty
git_version_to_python = git_version_to_python_version

__all__ = [
    "black",
    "BlackTask",
    "build",
    "BuildTask",
    "flake8",
    "Flake8Task",
    "git_version_to_python_version",
    "git_version_to_python",
    "install",
    "InstallTask",
    "InfoTask",
    "info",
    "isort",
    "IsortTask",
    "login",
    "mypy",
    "MypyTask",
    "mypy_stubtest",
    "MypyStubtestTask",
    "publish",
    "PublishTask",
    "pycln",
    "PyclnTask",
    "pylint",
    "PylintTask",
    "pytest",
    "PytestTask",
    "CoverageFormat",
    "python_settings",
    "PythonSettings",
    "pyupgrade",
    "PyUpgradeTask",
    "PyUpgradeCheckTask",
    "update_lockfile_task",
    "update_pyproject_task",
]



from __future__ import annotations

import logging
from pathlib import Path

from kraken.common import Supplier
from kraken.core import Project, Property, Task, TaskStatus

from ..buildsystem import PythonBuildSystem
from ..settings import python_settings

logger = logging.getLogger(__name__)


class BuildTask(Task):
    build_system: Property[PythonBuildSystem | None]
    output_directory: Property[Path]
    as_version: Property[str | None] = Property.default(None)
    output_files: Property[list[Path]] = Property.output()

    # Task

    def get_description(self) -> str | None:
        build_system = self.build_system.get()
        return (
            f"Build distributions for your Python project. [build system: "
            f"{build_system.name if build_system else None}]"
        )

    def execute(self) -> TaskStatus:
        build_system = self.build_system.get()
        if not build_system:
            return TaskStatus.failed("no build system configured")
        output_directory = self.output_directory.get_or(self.project.build_directory / "python-dist")
        output_directory.mkdir(exist_ok=True, parents=True)
        self.output_files.set(build_system.build(output_directory, self.as_version.get()))
        return TaskStatus.succeeded()


def build(
    *,
    name: str = "python.build",
    group: str | None = "build",
    as_version: str | None = None,
    project: Project | None = None,
) -> BuildTask:
    """Creates a build task for the given project.

    The build task relies on the build system configured in the Python project settings."""

    project = project or Project.current()
    task = project.task(name, BuildTask, group=group)
    task.build_system = Supplier.of_callable(lambda: python_settings(project).build_system)
    task.as_version = as_version
    return task



from __future__ import annotations

import contextlib
from pathlib import Path
from typing import Any

from kraken.common import atomic_file_swap
from kraken.core import BackgroundTask, Property, TaskStatus

from kraken.std.cargo import CargoProject
from kraken.std.cargo.manifest import CargoManifest


class CargoBumpVersionTask(BackgroundTask):
    """This task bumps the version numbers in `Cargo.toml`, and if a registry is specified, updates the registry and
    version of 'path' dependencies. The change can be reverted afterwards if the :attr:`revert` option is enabled."""

    description = 'Bump the version in "%(cargo_toml_file)s" to "%(version)s" [temporary: %(revert)s]'
    version: Property[str]
    registry: Property[str]
    revert: Property[bool] = Property.default(False)
    cargo_toml_file: Property[Path]

    def _get_updated_cargo_toml(self) -> str:
        manifest = CargoManifest.read(self.cargo_toml_file.get())
        if manifest.package is None:
            return manifest.to_toml_string()

        # Cargo does not play nicely with semver metadata (ie. 1.0.1-dev3+abc123)
        # We replace that to 1.0.1-dev3abc123
        fixed_version_string = self.version.get().replace("+", "")
        manifest.package.version = fixed_version_string
        if manifest.workspace and manifest.workspace.package:
            manifest.workspace.package.version = self.version.get()

        if self.registry.is_filled():
            cargo = CargoProject.get_or_create(self.project)
            registry = cargo.registries[self.registry.get()]
            if manifest.dependencies:
                self._push_version_to_path_deps(fixed_version_string, manifest.dependencies.data, registry.alias)
            if manifest.build_dependencies:
                self._push_version_to_path_deps(fixed_version_string, manifest.build_dependencies.data, registry.alias)
        return manifest.to_toml_string()

    def _push_version_to_path_deps(
        self, version_string: str, dependencies: dict[str, Any], registry_alias: str
    ) -> None:
        """For each dependency in the given dependencies, if the dependency is a `path` dependency, injects the current
        version and registry (required for publishing - path dependencies cannot be published alone)."""
        for dep_name in dependencies:
            dependency = dependencies[dep_name]
            if isinstance(dependency, dict):
                if "path" in dependency:
                    dependency["version"] = f"={version_string}"
                    dependency["registry"] = registry_alias

    # BackgroundTask

    def start_background_task(self, exit_stack: contextlib.ExitStack) -> TaskStatus | None:
        content = self._get_updated_cargo_toml()
        revert = self.revert.get()
        fp = exit_stack.enter_context(atomic_file_swap(self.cargo_toml_file.get(), "w", always_revert=revert))
        fp.write(content)
        fp.close()
        version = self.version.get()
        return (
            TaskStatus.started(f"temporary bump to {version}")
            if revert
            else TaskStatus.succeeded(f"permanent bump to {version}")
        )



from __future__ import annotations

import logging
from pathlib import Path
from typing import Literal

from kraken.core import Project, Property

from kraken.std.git.gitignore.generated import join_generated_section, split_generated_section
from kraken.std.util.render_file_task import RenderFileTask

from ..gitignore.gitignore_io import gitignore_io_fetch_cached
from ..gitignore.parser import GitignoreEntry, GitignoreFile

logger = logging.getLogger(__name__)


class GitignoreSyncTask(RenderFileTask):
    """
    This task ensures that a `.gitignore` file contains the generated content that it is expected to contain,
    in addition to preserving any user-defined content.
    """

    #: The generated content that the `.gitignore` file is expected to contain. This content will be
    #: enclosed in markers that indicate the start and end of the generated section.
    generated_content: Property[list[str]] = Property.default_factory(
        lambda: [
            "# Kraken",
            "/build",
        ]
    )

    #: Specify a list of gitignore.io tokens to include in addition to #generated_content. These tokens
    #: will be placed before the values from #generated_content.
    gitignore_io_tokens: Property[list[str]] = Property.default_factory(list)

    #: Whether to permit backfilling gitignore.io tokens that are not distributed with kraken-std via
    #: and HTTP request to gitignore.io.
    gitignore_io_allow_http_request_backfill: Property[bool] = Property.default(False)

    #: Where to place the generated content, if no generated content is found in the file.
    where: Property[Literal["top", "bottom"]] = Property.default("top")

    def get_file_contents(self, file: Path) -> str:
        if file.exists():
            # Read the existing content, so we can replace the generated section.
            user1, generated, user2 = split_generated_section(GitignoreFile.parse(file))
        else:
            user1, generated, user2 = GitignoreFile(), GitignoreFile(), GitignoreFile()

        # When no generated section is found, the entire file content sits in user1. If we want our new generated
        # content to sit at the top, we need to swap user1 and user2.
        if not generated and self.where.get() == "top":
            user1, user2 = user2, user1

        # Replace the generated content.
        generated = GitignoreFile.parse(self.generated_content.get())
        if tokens := self.gitignore_io_tokens.get():
            generated += GitignoreFile.parse(
                gitignore_io_fetch_cached(tokens, backfill=self.gitignore_io_allow_http_request_backfill.get())
            )

        # Ensure there's at least one blank space between sections.
        if user1 and generated and (user1[-1] != "" and generated[0] != ""):
            user1.append(GitignoreEntry())
        if generated and user2 and (generated[-1] != "" and user2[0] != ""):
            user2.insert(0, GitignoreEntry())

        return "\n".join(join_generated_section(user1, generated, user2)) + "\n"

    # RenderFileTask

    #: The path to the `.gitignore` file to sync, relative to the task's project directory.
    file: Property[Path] = Property.default(Path(".gitignore"))

    # Task

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.content.setcallable(lambda: self.get_file_contents(self.file.get()))

    def finalize(self) -> None:
        self.file.setmap(lambda file: self.project.directory / file)
        super().finalize()



from __future__ import annotations

from pathlib import Path

from kraken.common.path import try_relative_to
from kraken.common.strings import as_bytes
from kraken.common.supplier import Supplier
from kraken.core import Project, Property, Task, TaskStatus

from .check_file_contents_task import CheckFileContentsTask

DEFAULT_ENCODING = "utf-8"


class RenderFileTask(Task):
    """The RenderFileTask renders a single file to disk.

    The contents of the file can be provided by the :attr:`content` property or by creating a subclass
    that implements the :meth:`get_file_contents` method.

    It is common for a RenderFileTask to be added to the default `apply` task group. A matching check task,
    should you create it, would be a good candidate to add to the default `check` group.
    """

    description = 'Create or update "%(file)s".'

    file: Property[Path]
    content: Property[str | bytes]
    encoding: Property[str] = Property.default(DEFAULT_ENCODING)

    def create_check(
        self,
        name: str = "{name}.check",
        task_class: type[CheckFileContentsTask] | None = None,
        description: str | None = None,
        group: str | None = "check",
    ) -> CheckFileContentsTask:
        task = self.project.task(
            name.replace("{name}", self.name), task_class or CheckFileContentsTask, description=description, group=group
        )
        task.file = self.file.value
        task.content = self.content.value
        task.encoding = self.encoding.value
        task.render_prepare = Supplier.of_callable(self.prepare)
        task.depends_on(self, mode="order-only")
        return task

    # Task

    def prepare(self) -> TaskStatus:
        file = self.file.get()
        if file.is_file() and file.read_bytes() == as_bytes(self.content.get(), self.encoding.get()):
            return TaskStatus.up_to_date(f'"{try_relative_to(file)}" is up to date')
        return TaskStatus.pending()

    def execute(self) -> TaskStatus:
        file = self.file.get()
        file.parent.mkdir(exist_ok=True, parents=True)
        content = as_bytes(self.content.get(), self.encoding.get())
        file.write_bytes(content)
        return TaskStatus.succeeded(f"wrote {len(content)} bytes to {try_relative_to(file)}")


def render_file(
    name: str,
    description: str | None = None,
    group: str | None = "apply",
    create_check: bool = True,
    check_name: str = "{name}.check",
    check_group: str | None = "check",
    check_description: str | None = None,
    project: Project | None = None,
    task_class: type[RenderFileTask] | None = None,
    check_task_class: type[CheckFileContentsTask] | None = None,
    *,
    file: str | Path | Supplier[Path],
    content: str | Supplier[str],
    encoding: str | Supplier[str] = DEFAULT_ENCODING,
) -> tuple[RenderFileTask, CheckFileContentsTask | None]:
    project = project or Project.current()
    render_task = project.task(name, task_class or RenderFileTask, description=description, group=group)
    render_task.file = Path(file) if isinstance(file, str) else file
    render_task.content = content
    render_task.encoding = encoding

    if create_check:
        check_task = render_task.create_check(
            check_name.replace("{name}", name), check_task_class, description=check_description, group=check_group
        )
    else:
        check_task = None

    return render_task, check_task



from __future__ import annotations

from pathlib import Path

from kraken.core import Property, TaskStatus

from ._cargo_sqlx import CargoBaseSqlxTask


class CargoSqlxPrepareTask(CargoBaseSqlxTask):
    """Using cargo sqlx, generate the sqlx-data.json file for offline mode. If check=True, will verify that the
    sqlx-data.json file is up-to-date with the current database schema and code queries."""

    migrations: Property[Path]
    check: Property[bool] = Property.default(False)

    def execute(self) -> TaskStatus:
        arguments = ["prepare"]
        if self.check.get():
            arguments.append("--check")

        return self._execute_command(arguments)

    def get_description(self) -> str | None:
        if self.check.get():
            return "Check that sqlx-data.json file is up-to-date with the current database schema and code queries"
        return "Generate the sqlx-data.json file for offline mode"



""" General utility tasks. """

from __future__ import annotations

from .copyright_task import check_and_format_copyright

__all__ = ["check_file_exists_and_is_committed", "check_valid_readme_exists", "check_and_format_copyright"]



import os
import unittest.mock

from kraken.core import Project
from kraken.core.system.task import TaskStatus
from pytest import CaptureFixture

from kraken.std.util.check_file_contents_task import CheckFileContentsTask


@unittest.mock.patch.dict(os.environ, {"ANSI_COLORS_DISABLED": "1"})
def test__CheckFileContentsTask__shows_diff(kraken_project: Project, capsys: CaptureFixture[str]) -> None:
    path = kraken_project.directory / "file.txt"
    path.write_text("Hello, world!\n", encoding="utf8")

    task = kraken_project.task("checkFile", CheckFileContentsTask)
    task.file = path
    task.encoding = "utf8"
    task.content = "Hello, world!\nGoodbye, world!\n"
    status = task.execute()
    assert status == TaskStatus.failed(f'file "{path}" is not up to date')

    assert capsys.readouterr().out.splitlines(keepends=True) == [
        "  Hello, world!\n",
        "+ Goodbye, world!\n",
    ]



from base64 import b64encode
from json import loads

from mitmproxy import addonmanager, ctx, http


class AuthInjector:
    def __init__(self) -> None:
        self.auth: dict[str, tuple[str, str]] = {}

    def load(self, loader: addonmanager.Loader) -> None:
        loader.add_option(
            name="auth",
            typespec=str,
            help="The auth mapping (host to tuple of (username, password)).",
            default="{}",
        )

    def request(self, flow: http.HTTPFlow) -> None:
        auth = loads(ctx.options.auth)
        if "Authorization" not in flow.request.headers and flow.request.host in auth:
            creds = auth[flow.request.host]
            auth_payload = b64encode(f"{creds[0]}:{creds[1]}".encode()).decode()
            flow.request.headers["Authorization"] = "Basic " + auth_payload


addons = [AuthInjector()]



from __future__ import annotations

import logging
import shutil
import subprocess as sp

from kraken.core import Property, Task, TaskStatus

from kraken.std.cargo.config import CargoConfig

logger = logging.getLogger(__name__)


class CargoFmtTask(Task):
    check: Property[bool] = Property.default(False)
    all_packages: Property[bool] = Property.default(False)
    config: Property[CargoConfig]

    def execute(self) -> TaskStatus:
        nightly = self.config.get().nightly
        if nightly:
            result = self.check_nightly_toolchain()
            if result.is_not_ok():
                return result

        command = self.get_command()

        return TaskStatus.from_exit_code(command, sp.call(command, cwd=self.project.directory))

    def get_command(self) -> list[str]:
        command = ["cargo"]

        if self.config.get().nightly:
            command.append("+nightly")

        command.append("fmt")

        if self.check.get():
            command.append("--check")
        if self.all_packages.get():
            command.append("--all")

        return command

    def check_nightly_toolchain(self) -> TaskStatus:
        if CargoFmtTask.is_using_rust_nightly():
            # if a nightly version of rust is used, then it is in the toolchains
            return TaskStatus.succeeded()

        # if the current version of rust is not nightly, we try to find a nightly toolchain with the correct
        # architecture
        # this is a best effort check to return a helpful error message by checking if available toolchains, which is
        # only possible if rustup is installed
        rustup = shutil.which("rustup")
        if rustup is None:
            return TaskStatus.succeeded()

        result = sp.run([rustup, "show", "active-toolchain"], stdout=sp.PIPE, stderr=sp.DEVNULL)
        if result.returncode != 0:
            return TaskStatus.failed("could not determine the active rust toolchain")

        output = result.stdout.decode("utf-8").split(" ")
        architecture = "-".join(output[0].split("-")[-3:])

        result = sp.run([rustup, "show"], stdout=sp.PIPE, stderr=sp.DEVNULL)
        if result.returncode != 0:
            return TaskStatus.failed("could not determine available toolchains")

        if any(
            [
                toolchain.startswith("nightly-") and toolchain.endswith(architecture)
                for toolchain in result.stdout.decode("utf-8").splitlines()
            ]
        ):
            return TaskStatus.succeeded()

        formatted_command = " ".join(self.get_command())
        return TaskStatus.failed(
            f"cannot run `{formatted_command}` with nightly toolchain because it is not installed, either install it or"
            + " opt-out of the nightly toolchain with `cargo_config(nightly=False)`"
        )

    @staticmethod
    def is_using_rust_nightly() -> bool:
        rustc = shutil.which("rustc")
        if rustc is None:
            logger.warn("rustc is not installed, could not determine the active rust version")
            return False

        result = sp.run([rustc, "--version"], stdout=sp.PIPE, stderr=sp.DEVNULL)
        if result.returncode != 0:
            logger.warn("rustc --version is not installed, could not determine the active rust version")
            return False

        return "-nightly" in result.stdout.decode("utf-8")

    def get_description(self) -> str | None:
        formatted_command = " ".join(self.get_command())
        return f"Run `{formatted_command}`."



import hashlib

from kraken.core import Project
from kraken.core.system.task import TaskStatus

from kraken.std.util.validate_readme import ValidateReadmeTask

EXAMPLE_README = """
# project name

This readme is auto generated.
""".lstrip()

BAD_CONTENT_MD5_SUM = hashlib.md5(b"This readme is auto generated.\n").hexdigest()


def _write_example_readme(kraken_project: Project) -> None:
    (kraken_project.directory / "README.md").write_text(EXAMPLE_README)


def test__ValidateReadmeTask__fails_on_no_preferred_readme_file(kraken_project: Project) -> None:
    _write_example_readme(kraken_project)
    task = kraken_project.task("validate_readme", ValidateReadmeTask)
    task.preferred_file_names = ["README.rst"]
    status = task.execute()
    assert status == TaskStatus.failed("completed 1 out of 1 check(s) with errors")
    assert task.statuses.get() == [
        TaskStatus.failed("No preferred readme file found (expected one of: README.rst)"),
    ]


def test__ValidateReadmeTask__warns_on_disallowed_content(kraken_project: Project) -> None:
    _write_example_readme(kraken_project)
    task = kraken_project.task("validate_readme", ValidateReadmeTask)
    task.disallowed_md5_content_hashes = {"bad hash": BAD_CONTENT_MD5_SUM}
    status = task.execute()
    assert status == TaskStatus.warning("completed 1 out of 2 check(s) with warnings")
    assert task.statuses.get() == [
        TaskStatus.succeeded("Found readme file: README.md"),
        TaskStatus.warning("Readme file content matches disallowed hash 'bad hash'"),
    ]


def test__ValidateReadmeTask__warns_on_multiple_readme_files(kraken_project: Project) -> None:
    _write_example_readme(kraken_project)
    (kraken_project.directory / "README.rst").write_text("foobar")
    task = kraken_project.task("validate_readme", ValidateReadmeTask)
    status = task.execute()
    assert status == TaskStatus.warning("completed 1 out of 1 check(s) with warnings")
    assert task.statuses.get() == [
        TaskStatus.warning("Found multiple readme files. Linting only README.md"),
    ]


def test__ValidateReadmeTask__warns_about_disallowed_pattern_match(kraken_project: Project) -> None:
    _write_example_readme(kraken_project)
    task = kraken_project.task("validate_readme", ValidateReadmeTask)
    task.disallowed_regex_patterns = [r"auto\s*generated"]
    status = task.execute()
    assert status == TaskStatus.warning("completed 1 out of 2 check(s) with warnings")
    assert task.statuses.get() == [
        TaskStatus.succeeded("Found readme file: README.md"),
        TaskStatus.warning(r"Readme file contains disallowed pattern auto\s*generated"),
    ]


def test__ValidateReadmeTask__warns_about_minimum_lines(kraken_project: Project) -> None:
    _write_example_readme(kraken_project)
    task = kraken_project.task("validate_readme", ValidateReadmeTask)
    task.minimum_lines = 10
    status = task.execute()
    assert status == TaskStatus.warning("completed 1 out of 2 check(s) with warnings")
    assert task.statuses.get() == [
        TaskStatus.succeeded("Found readme file: README.md"),
        TaskStatus.warning(r"Readme file has less than 10 lines (got: 4)"),
    ]



""" Provides tasks for Rust projects that build using Cargo. """

from __future__ import annotations

import dataclasses
from collections.abc import Collection, Sequence
from pathlib import Path
from typing import Literal

from kraken.common import Supplier
from kraken.core import Project

from .config import CargoConfig, CargoProject, CargoRegistry
from .tasks.cargo_auth_proxy_task import CargoAuthProxyTask
from .tasks.cargo_build_task import CargoBuildTask
from .tasks.cargo_bump_version_task import CargoBumpVersionTask
from .tasks.cargo_check_toolchain_version import CargoCheckToolchainVersionTask
from .tasks.cargo_clippy_task import CargoClippyTask
from .tasks.cargo_deny_task import CargoDenyTask
from .tasks.cargo_fmt_task import CargoFmtTask
from .tasks.cargo_generate_deb import CargoGenerateDebPackage
from .tasks.cargo_publish_task import CargoPublishTask
from .tasks.cargo_sqlx_migrate import CargoSqlxMigrateTask
from .tasks.cargo_sqlx_prepare import CargoSqlxPrepareTask
from .tasks.cargo_sync_config_task import CargoSyncConfigTask
from .tasks.cargo_test_task import CargoTestTask
from .tasks.cargo_update_task import CargoUpdateTask
from .tasks.rustup_target_add_task import RustupTargetAddTask

__all__ = [
    "cargo_auth_proxy",
    "cargo_build",
    "cargo_bump_version",
    "cargo_clippy",
    "cargo_deny",
    "cargo_fmt",
    "cargo_generate_deb_package",
    "cargo_publish",
    "cargo_registry",
    "cargo_sqlx_migrate",
    "cargo_sqlx_prepare",
    "cargo_sync_config",
    "cargo_update",
    "CargoAuthProxyTask",
    "CargoBuildTask",
    "CargoBumpVersionTask",
    "CargoClippyTask",
    "CargoDenyTask",
    "CargoProject",
    "CargoPublishTask",
    "CargoRegistry",
    "CargoSqlxMigrateTask",
    "CargoSqlxPrepareTask",
    "CargoSyncConfigTask",
    "CargoTestTask",
    "cargo_check_toolchain_version",
    "CargoCheckToolchainVersionTask",
    "rustup_target_add",
    "RustupTargetAddTask",
]

#: This is the name of a group in every project that contains Cargo tasks to contain the tasks that either support
#: or establish pre-requisites for a Cargo build to be executed. This includes ensuring certain configuration is
#: is up to date and the Cargo auth proxy if it is being used.
CARGO_BUILD_SUPPORT_GROUP_NAME = "cargoBuildSupport"
#: This is the name of a group in every project that are pre-requisites for publishing crates within a Cargo workspace.
#: This includes ensuring that all path dependencies have up to date version numbers.
CARGO_PUBLISH_SUPPORT_GROUP_NAME = "cargoPublishSupport"


def cargo_config(*, project: Project | None = None, nightly: bool = False) -> CargoConfig:
    project = project or Project.current()
    config = CargoConfig(nightly=nightly)
    project.metadata.append(config)
    return config


def cargo_sqlx_migrate(
    *,
    name: str = "sqlxMigrate",
    project: Project | None = None,
    base_directory: Path | None = None,
    database_url: str | None = None,
    migrations: Path | None = None,
) -> CargoSqlxMigrateTask:
    project = project or Project.current()
    task = project.task(name, CargoSqlxMigrateTask)
    task.base_directory = base_directory
    task.database_url = database_url
    task.migrations = migrations
    return task


def cargo_sqlx_prepare(
    *,
    name: str = "sqlxPrepare",
    project: Project | None = None,
    check: bool,
    base_directory: Path | None = None,
    database_url: str | None = None,
    migrations: Path | None = None,
) -> CargoSqlxPrepareTask:
    project = project or Project.current()
    name = f"{name}Check" if check else name
    task = project.task(name, CargoSqlxPrepareTask, group="check" if check else None)
    task.check = check
    task.base_directory = base_directory
    task.database_url = database_url
    task.migrations = migrations

    # Preparing or checking sqlx metadata calls `cargo metadata`, which can require the auth proxy
    # Without the auth proxy, cargo sqlx commands would fail with a cryptic error
    # See https://github.com/launchbadge/sqlx/pull/2222 for details
    task.depends_on(f":{CARGO_BUILD_SUPPORT_GROUP_NAME}?")

    return task


def cargo_registry(
    alias: str,
    index: str,
    read_credentials: tuple[str, str] | None = None,
    publish_token: str | None = None,
    project: Project | None = None,
) -> None:
    """Adds a Cargo registry to the project. The registry must be synced to disk into the `.cargo/config.toml`
    configuration file. You need to make sure to add a sync task using :func:`cargo_sync_config` if you manage
    your Cargo registries with this function. Can be called multiple times.

    :param alias: The registry alias.
    :param index: The registry index URL (usually an HTTPS URL that ends in `.git`).
    :param read_credentials: Username/password to read from the registry (only for private registries).
    :param publish_token: The token to use with `cargo publish`.

    !!! note Artifactory

        It appears that for Artifactory, the *publish_token* must be of the form `Bearer <TOKEN>` where the token
        is a token generated manually via the JFrog UI. It cannot be an API key.
    """

    cargo = CargoProject.get_or_create(project)
    cargo.add_registry(alias, index, read_credentials, publish_token)


def cargo_auth_proxy(*, project: Project | None = None) -> CargoAuthProxyTask:
    """Creates a background task that the :func:`cargo_build` and :func:`cargo_publish` tasks will depend on to
    inject the read credentials for private registries into HTTPS requests made by Cargo. This is only needed when
    private registries are used."""

    project = project or Project.current()
    cargo = CargoProject.get_or_create(project)

    task = project.task("cargoAuthProxy", CargoAuthProxyTask, group=CARGO_BUILD_SUPPORT_GROUP_NAME)
    task.registries = Supplier.of_callable(lambda: list(cargo.registries.values()))

    # The auth proxy is required for both building and publishing cargo packages with private cargo project dependencies
    project.group(CARGO_PUBLISH_SUPPORT_GROUP_NAME).add(task)

    # The auth proxy injects values into the cargo config, the cargoSyncConfig.check ensures that it reflects
    # the temporary changes that should be made to the config. The check has to run before the auth proxy,
    # otheerwise it is garuanteed to fail.
    task.depends_on(":cargoSyncConfig.check?", mode="order-only")
    return task


def cargo_sync_config(
    *,
    replace: bool = False,
    project: Project | None = None,
) -> CargoSyncConfigTask:
    """Creates a task that the :func:`cargo_build` and :func:`cargo_publish` tasks will depend on to synchronize
    the `.cargo/config.toml` configuration file, ensuring that the Cargo registries configured with the
    :func:`cargo_registry` function are present and up to date."""

    project = project or Project.current()
    cargo = CargoProject.get_or_create(project)
    task = project.task("cargoSyncConfig", CargoSyncConfigTask, group="apply")
    task.registries = Supplier.of_callable(lambda: list(cargo.registries.values()))
    task.replace = replace
    check_task = task.create_check()
    project.group(CARGO_BUILD_SUPPORT_GROUP_NAME).add(check_task)
    return task


def cargo_clippy(
    *,
    allow: str = "staged",
    fix: bool = False,
    name: str | None = None,
    group: str | None = "_auto_",
    project: Project | None = None,
) -> CargoClippyTask:
    project = project or Project.current()
    name = "cargoClippyFix" if fix else "cargoClippy"
    group = ("fmt" if fix else "lint") if group == "_auto_" else group
    cargo = CargoProject.get_or_create(project)
    task = project.task(name, CargoClippyTask, group=group)
    task.fix = fix
    task.allow = allow
    task.env = Supplier.of_callable(lambda: cargo.build_env)

    # Clippy builds your code.
    task.depends_on(f":{CARGO_BUILD_SUPPORT_GROUP_NAME}?")

    return task


def cargo_deny(
    *,
    project: Project | None = None,
    checks: Sequence[str] | Supplier[Sequence[str]] = (),
    config_file: Path | Supplier[Path] | None = None,
    error_message: str | None = None,
) -> CargoDenyTask:
    """Adds a task running cargo-deny for cargo projects. This checks different rules on dependencies, such as scanning
    for vulnerabilities, unwanted licences, or custom bans.

    :param checks: The list of cargo-deny checks to run, as defined in
    https://embarkstudios.github.io/cargo-deny/checks/index.html. If not provided, defaults to all of them.
    :param config_file: The configuration file as defined in https://embarkstudios.github.io/cargo-deny/checks/cfg.html
    If not provided defaults to cargo-deny default location.
    :param error_message: The error message to show if the task fails.
    """

    project = project or Project.current()
    return project.do(
        "cargoDeny",
        CargoDenyTask,
        checks=checks,
        config_file=config_file,
        error_message=error_message,
    )


@dataclasses.dataclass
class CargoFmtTasks:
    check: CargoFmtTask
    format: CargoFmtTask


def cargo_fmt(*, all_packages: bool = False, project: Project | None = None) -> CargoFmtTasks:
    project = project or Project.current()
    config = project.find_metadata(CargoConfig) or cargo_config(project=project)
    format = project.task("cargoFmt", CargoFmtTask, group="fmt")
    format.all_packages = all_packages
    format.config = config

    check = project.task("cargoFmtCheck", CargoFmtTask, group="lint")
    check.all_packages = all_packages
    check.config = config
    check.check = True
    return CargoFmtTasks(check=check, format=format)


def cargo_update(*, project: Project | None = None) -> CargoUpdateTask:
    project = project or Project.current()
    task = project.task("cargoUpdate", CargoUpdateTask, group="update")
    task.depends_on(":cargoBuildSupport")
    return task


def cargo_bump_version(
    *,
    version: str,
    revert: bool = True,
    name: str = "cargoBumpVersion",
    registry: str | None = None,
    project: Project | None = None,
    cargo_toml_file: Path = Path("Cargo.toml"),
) -> CargoBumpVersionTask:
    """Get or create a task that bumps the version in `Cargo.toml`.

    :param version: The version number to bump to.
    :param revert: Revert the version number after all direct dependants have run.
    :param name: The task name. Note that if another task with the same configuration but different name exists,
        it will not change the name of the task and that task will still be reused.
    :param group: The group to assign the task to (even if the task is reused)."""

    # This task will write the "current" version (usually taken from git) into Cargo.toml
    # That is useful at two places
    #  * at build time (because apps may use env!("CARGO_PKG_VERSION"))
    #  * at publish time (because the artifact file name is derived from the version)

    project = project or Project.current()

    task = project.task(name, CargoBumpVersionTask, group=CARGO_BUILD_SUPPORT_GROUP_NAME)
    task.version = version
    task.revert = revert
    task.registry = registry
    task.cargo_toml_file = cargo_toml_file

    # project.do() only supports one group, but this task is needed in two groups
    project.group(CARGO_PUBLISH_SUPPORT_GROUP_NAME).add(task)

    return task


def cargo_build(
    mode: Literal["debug", "release"],
    incremental: bool | None = None,
    env: dict[str, str] | None = None,
    workspace: bool = False,
    *,
    exclude: Collection[str] = (),
    group: str | None = "build",
    name: str | None = None,
    project: Project | None = None,
    features: list[str] | None = None,
) -> CargoBuildTask:
    """Creates a task that runs `cargo build`.

    :param mode: Whether to create a task that runs the debug or release build.
    :param incremental: Whether to build incrementally or not (with the `--incremental=` option). If not
        specified, the option is not specified and the default behaviour is used.
    :param env: Override variables for the build environment variables. Values in this dictionary override
        variables in :attr:`CargoProject.build_env`.
    :param exclude: List of workspace crates to exclude from the build.
    :param name: The name of the task. If not specified, defaults to `:cargoBuild{mode.capitalised()}`.
    :param features: List of Cargo features to enable in the build."""

    assert mode in ("debug", "release"), repr(mode)
    project = project or Project.current()
    cargo = CargoProject.get_or_create(project)

    additional_args = []
    if workspace:
        additional_args.append("--workspace")
    for crate in exclude:
        additional_args.append("--exclude")
        additional_args.append(crate)
    if mode == "release":
        additional_args.append("--release")
    if features:
        additional_args.append("--features")
        # `cargo build` expects features to be comma separated, in one string.
        # For example `cargo build --features abc,efg` instead of `cargo build --features abc efg`.
        additional_args.append(",".join(features))

    task = project.task(
        f"cargoBuild{mode.capitalize()}" if name is None else name,
        CargoBuildTask,
        group=group,
    )
    task.incremental = incremental
    task.target = mode
    task.additional_args = additional_args
    task.env = Supplier.of_callable(lambda: {**cargo.build_env, **(env or {})})

    task.depends_on(f":{CARGO_BUILD_SUPPORT_GROUP_NAME}?")
    return task


def cargo_test(
    incremental: bool | None = None,
    env: dict[str, str] | None = None,
    *,
    group: str | None = "test",
    project: Project | None = None,
    features: list[str] | None = None,
) -> CargoTestTask:
    """Creates a task that runs `cargo test`.

    :param incremental: Whether to build the tests incrementally or not (with the `--incremental=` option). If not
        specified, the option is not specified and the default behaviour is used.
    :param env: Override variables for the build environment variables. Values in this dictionary override
        variables in :attr:`CargoProject.build_env`.
    :param features: List of Cargo features to enable in the build."""

    project = project or Project.current()
    cargo = CargoProject.get_or_create(project)

    additional_args = []
    if features:
        additional_args.append("--features")
        # `cargo build` expects features to be comma separated, in one string.
        # for example `cargo build --features abc,efg` instead of `cargo build --features abc efg`.
        additional_args.append(",".join(features))

    task = project.task("cargoTest", CargoTestTask, group=group)
    task.incremental = incremental
    task.additional_args = additional_args
    task.env = Supplier.of_callable(lambda: {**cargo.build_env, **(env or {})})
    task.depends_on(f":{CARGO_BUILD_SUPPORT_GROUP_NAME}?")
    return task


def cargo_publish(
    registry: str,
    incremental: bool | None = None,
    env: dict[str, str] | None = None,
    *,
    verify: bool = True,
    retry_attempts: int = 0,
    additional_args: Sequence[str] = (),
    name: str = "cargoPublish",
    package_name: str | None = None,
    project: Project | None = None,
) -> CargoPublishTask:
    """Creates a task that publishes the create to the specified *registry*.

    :param registry: The alias of the registry to publish to.
    :param incremental: Incremental builds on or off.
    :param env: Environment variables (overrides :attr:`CargoProject.build_env`).
    :param verify: If this is enabled, the `cargo publish` task will build the crate after it is packaged.
        Disabling this just packages the crate and publishes it. Only if this is enabled will the created
        task depend on the auth proxy.
    :param retry_attempts: Retry the publish task if it fails, up to a maximum number of attempts. Sometimes
        cargo publishes can be flakey depending on the destination. Defaults to 0 retries.
    """

    project = project or Project.current()
    cargo = CargoProject.get_or_create(project)

    task = project.task(
        f"{name}/{package_name}" if package_name is not None else name,
        CargoPublishTask,
        group="publish",
    )
    task.registry = Supplier.of_callable(lambda: cargo.registries[registry])
    task.additional_args = list(additional_args)
    task.allow_dirty = True
    task.incremental = incremental
    task.verify = verify
    task.retry_attempts = retry_attempts
    task.package_name = package_name
    task.env = Supplier.of_callable(lambda: {**cargo.build_env, **(env or {})})
    task.depends_on(f":{CARGO_PUBLISH_SUPPORT_GROUP_NAME}?")
    return task


def cargo_check_toolchain_version(
    minimal_version: str, *, project: Project | None = None
) -> CargoCheckToolchainVersionTask:
    """Creates a task that checks that cargo is at least at version `minimal_version`"""

    project = project or Project.current()
    task = project.task(
        f"cargoCheckVersion/{minimal_version}",
        CargoCheckToolchainVersionTask,
        group=CARGO_BUILD_SUPPORT_GROUP_NAME,
    )
    task.minimal_version = minimal_version
    return task


def rustup_target_add(target: str, *, group: str | None = None, project: Project | None = None) -> RustupTargetAddTask:
    """Creates a task that installs a given target for Cargo"""

    project = project or Project.current()
    task = project.task(f"rustupTargetAdd/{target}", RustupTargetAddTask, group=group)
    task.target = target
    return task


def cargo_generate_deb_package(*, project: Project | None = None, package_name: str) -> CargoGenerateDebPackage:
    project = project or Project.current()
    return project.do(
        "cargoGenerateDeb",
        CargoGenerateDebPackage,
        package_name=package_name,
    )



import os
import subprocess
from pathlib import Path

from kraken.core import Project, TaskStatus

from kraken.std.git.tasks.check_file_task import CheckFileTask


def _git_env() -> dict[str, str]:
    # We need to make sure we don't read any custom user config. For example, if the user running the tests has set up
    # commit signing, they may be asked for their GPG passphrase, which would block the tests.
    return {**os.environ, "GIT_CONFIG_NOSYSTEM": "true", "GIT_CONFIG": "noexist"}


def _git_init(path: Path) -> None:
    subprocess.run(["git", "init"], check=True, cwd=path, env=_git_env())


def _git_add(path: Path, files: list[Path]) -> None:
    subprocess.run(["git", "add", *map(str, files)], check=True, cwd=path, env=_git_env())


def _git_commit(path: Path, message: str) -> None:
    subprocess.run(
        ["git", "-c", "user.email=john@doe.com", "-c", "user.name=John Doe", "commit", "--allow-empty", "-m", message],
        check=True,
        cwd=path,
        env=_git_env(),
    )


def test__CheckFileTask__file_does_not_exist(kraken_project: Project) -> None:
    task = kraken_project.task("check_file", CheckFileTask)
    task.file_to_check = Path("warbl.garbl")
    status = task.execute()
    assert status == TaskStatus.failed("'warbl.garbl' does not exist")

    task.state.set("absent")
    status = task.execute()
    assert status == TaskStatus.succeeded("'warbl.garbl' does not exist")


def test__CheckFileTask__file_exists_but_is_not_committed(kraken_project: Project) -> None:
    _git_init(kraken_project.directory)
    _git_commit(kraken_project.directory, "Initial commit")
    file_name = Path("warbl.garbl")
    (kraken_project.directory / file_name).touch()

    task = kraken_project.task("check_file", CheckFileTask)
    task.file_to_check = file_name
    status = task.execute()
    assert status == TaskStatus.failed("'warbl.garbl' exists but is not committed")

    task.state.set("absent")
    assert status == TaskStatus.failed("'warbl.garbl' exists but is not committed")


def test__file_exists_and_is_committed(kraken_project: Project) -> None:
    _git_init(kraken_project.directory)
    file_name = Path("warbl.garbl")
    (kraken_project.directory / file_name).touch()
    _git_add(kraken_project.directory, [file_name])
    _git_commit(kraken_project.directory, "Initial commit")

    task = kraken_project.task("check_file", CheckFileTask)
    task.file_to_check = file_name
    status = task.execute()
    assert status == TaskStatus.succeeded("'warbl.garbl' exists and is committed")

    task.state.set("absent")
    status = task.execute()
    assert status == TaskStatus.failed("'warbl.garbl' exists and is committed")



from .cargo_build_task import CargoBuildTask


class CargoTestTask(CargoBuildTask):
    """This task runs `cargo test` using the specified parameters. It will respect the authentication
    credentials configured in :attr:`CargoProjectSettings.auth`."""

    description = "Run `cargo test`."

    def get_cargo_command(self, env: dict[str, str]) -> list[str]:
        super().get_cargo_command(env)
        return ["cargo", "test"] + self.additional_args.get()



from __future__ import annotations

from pathlib import Path
from typing import Literal

import tomli
import tomli_w
from kraken.core import Project, Property

from kraken.std.util.render_file_task import RenderFileTask

from ..config import CargoRegistry


class CargoSyncConfigTask(RenderFileTask):
    """This task updates the `.cargo/config.toml` file to inject configuration values."""

    file: Property[Path]

    #: If enabled, the configuration file will be replaced rather than updated.
    replace: Property[bool] = Property.default(False)

    #: The registries to insert into the configuration.
    registries: Property[list[CargoRegistry]] = Property.default_factory(list)

    #: Enable fetching Cargo indexes with the Git CLI.
    git_fetch_with_cli: Property[bool]

    #: Whether to use the sparse protocol for crates.io.
    crates_io_protocol: Property[Literal["git", "sparse"]] = Property.default("sparse")

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.file.setcallable(lambda: project.directory / ".cargo" / "config.toml")
        self.content.setcallable(lambda: self.get_file_contents(self.file.get()))

    def get_file_contents(self, file: Path) -> str | bytes:
        content = tomli.loads(file.read_text()) if not self.replace.get() and file.exists() else {}
        content.setdefault("registries", {})["crates-io"] = {"protocol": self.crates_io_protocol.get()}
        for registry in self.registries.get():
            content.setdefault("registries", {})[registry.alias] = {"index": registry.index}
        if self.git_fetch_with_cli.is_filled():
            if self.git_fetch_with_cli.get():
                content.setdefault("net", {})["git-fetch-with-cli"] = True
            else:
                if "net" in content:
                    content["net"].pop("git-fetch-with-cli", None)
        lines = []
        if self.replace.get():
            lines.append("# This file is managed by Kraken. Manual edits to this file will be overwritten.")
        else:
            lines.append(
                "# This file is partially managed by Kraken. Comments and manually added "
                "repositories are not preserved."
            )
        lines.append(tomli_w.dumps(content))
        return "\n".join(lines)



from __future__ import annotations

import dataclasses
from collections.abc import Sequence
from pathlib import Path

from kraken.core import Project, Property

from kraken.std.python.tasks.base_task import EnvironmentAwareDispatchTask


class CopyrightTask(EnvironmentAwareDispatchTask):
    """A task to run the `pyaddlicense` command to ensure copyright files exist in all source files."""

    python_dependencies = ["pyaddlicense"]

    check_only: Property[bool] = Property.default(False)
    holder: Property[str] = Property.default("")
    ignore: Property[Sequence[str]] = Property.default_factory(list)
    custom_license: Property[str]
    custom_license_file: Property[Path]

    def get_execute_command(self) -> list[str]:
        command = ["pyaddlicense"]

        if self.holder.get() != "":
            command += ["-o", f"{self.holder.get()}"]

        if self.check_only.get():
            command += ["-c"]

        if self.custom_license.is_filled():
            command += ["-l", f"{str(self.custom_license.get())}"]

        if self.custom_license_file.is_filled():
            command += ["-f", str(self.custom_license_file.get().absolute())]

        for custom_ignore in self.ignore.get():
            command += ["-i", custom_ignore]

        return command

    def get_description(self) -> str | None:
        if self.check_only.get():
            return "Check all source files start with a copyright comment via addlicense."
        else:
            return (
                "Check all source files start with a copyright message and edit them to include it if they do "
                "not (via addlicense)."
            )


@dataclasses.dataclass
class CopyrightTasks:
    check: CopyrightTask
    format: CopyrightTask


def check_and_format_copyright(
    holder: str,
    name: str = "copyright",
    project: Project | None = None,
    ignore: Sequence[str] = (),
    custom_license: str | None = None,
    custom_license_file: Path | None = None,
) -> CopyrightTasks:
    """Creates two copyright tasks, one to check and another to format. The check task will be grouped under `"lint"`
    whereas the format task will be grouped under `"fmt"`."""

    project = project or Project.current()
    check_task = project.task(f"{name}.check", CopyrightTask, group="lint")
    check_task.check_only = True
    check_task.holder = holder
    check_task.ignore = ignore
    check_task.custom_license = custom_license
    check_task.custom_license_file = custom_license_file

    format_task = project.task(f"{name}.fmt", CopyrightTask, group="fmt")
    format_task.check_only = False
    format_task.holder = holder
    format_task.ignore = ignore
    format_task.custom_license = custom_license
    format_task.custom_license_file = custom_license_file

    return CopyrightTasks(check_task, format_task)



# Perform HTTP requests
# This module internally calls httpx, but with a custom setup

import ssl
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any

import httpx

__all__ = ["request", "get", "options", "head", "post", "put", "patch", "delete", "stream"]

_CACHED_SYSTEM_CA_LIST: ssl.SSLContext | None = None


# Caching calls to ssl.create_default_context()
# Using functools.cache does not play well with mypy (https://github.com/python/mypy/issues/5107),
# so I'm implementing my own cache
def _get_system_ca_list() -> ssl.SSLContext:
    """
    Corporate proxies may replace the original SSL cert by their own.
    In this case, their cert should be part of the system list of accepted root CAs, and this is the one we
    should use, instead of httpx' default (which otherwise uses the list provided by certifi)
    """
    global _CACHED_SYSTEM_CA_LIST
    if _CACHED_SYSTEM_CA_LIST is None:
        _CACHED_SYSTEM_CA_LIST = ssl.create_default_context()
    return _CACHED_SYSTEM_CA_LIST


def request(method: str, url: str, **kwargs: Any) -> httpx.Response:
    return httpx.request(method, url, verify=_get_system_ca_list(), **kwargs)


def get(url: str, **kwargs: Any) -> httpx.Response:
    return httpx.get(url, verify=_get_system_ca_list(), **kwargs)


def options(url: str, **kwargs: Any) -> httpx.Response:
    return httpx.options(url, verify=_get_system_ca_list(), **kwargs)


def head(url: str, **kwargs: Any) -> httpx.Response:
    return httpx.head(url, verify=_get_system_ca_list(), **kwargs)


def post(url: str, **kwargs: Any) -> httpx.Response:
    return httpx.post(url, verify=_get_system_ca_list(), **kwargs)


def put(url: str, **kwargs: Any) -> httpx.Response:
    return httpx.put(url, verify=_get_system_ca_list(), **kwargs)


def patch(url: str, **kwargs: Any) -> httpx.Response:
    return httpx.patch(url, verify=_get_system_ca_list(), **kwargs)


def delete(url: str, **kwargs: Any) -> httpx.Response:
    return httpx.delete(url, verify=_get_system_ca_list(), **kwargs)


# Forwarding calls to `stream` would not be accepted as-is by mypy.
# Let's trick a bit
@contextmanager
def stream(url: str, **kwargs: Any) -> Iterator[httpx.Response]:
    with httpx.stream(url, verify=_get_system_ca_list(), **kwargs) as stream:
        yield stream



from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path

from kraken.common import Supplier
from kraken.core import Project, Property

from .base_task import EnvironmentAwareDispatchTask


class MypyTask(EnvironmentAwareDispatchTask):
    description = "Static type checking for Python code using Mypy."
    python_dependencies = ["mypy"]

    config_file: Property[Path]
    additional_args: Property[Sequence[str]] = Property.default_factory(list)
    check_tests: Property[bool] = Property.default(True)
    use_daemon: Property[bool] = Property.default(True)
    python_version: Property[str]

    # EnvironmentAwareDispatchTask

    def get_execute_command(self) -> list[str]:
        # TODO (@NiklasRosenstein): Should we somewhere add a task that ensures `.dmypy.json` is in `.gitignore`?
        #       Having it in the project directory makes it easier to just stop the daemon if it malfunctions (which
        #       happens regularly but is hard to detect automatically).
        status_file = (self.project.directory / ".dmypy.json").absolute()
        command = ["dmypy", "--status-file", str(status_file), "run", "--"] if self.use_daemon.get() else ["mypy"]
        if self.config_file.is_filled():
            command += ["--config-file", str(self.config_file.get().absolute())]
        else:
            command += ["--show-error-codes", "--namespace-packages"]  # Sane defaults. 🙏
        if self.python_version.is_filled():
            command += ["--python-version", self.python_version.get()]
        source_dir = self.settings.source_directory
        command += [str(source_dir)]
        if self.check_tests.get():
            # We only want to add the tests directory if it is not already in the source directory. Otherwise
            # Mypy will find the test files twice and error.
            tests_dir = self.settings.get_tests_directory()
            if tests_dir:
                try:
                    tests_dir.relative_to(source_dir)
                except ValueError:
                    command += [str(tests_dir)]
        command += self.additional_args.get()
        return command


def mypy(
    *,
    name: str = "python.mypy",
    project: Project | None = None,
    config_file: Path | Supplier[Path] | None = None,
    additional_args: Sequence[str] | Supplier[Sequence[str]] = (),
    check_tests: bool = True,
    use_daemon: bool = True,
    python_version: str | Supplier[str] | None = None,
) -> MypyTask:
    project = project or Project.current()
    task = project.task(name, MypyTask, group="lint")
    task.config_file = config_file
    task.additional_args = additional_args
    task.check_tests = check_tests
    task.use_daemon = use_daemon
    task.python_version = python_version
    return task



from __future__ import annotations

import enum
import os
import shlex
from collections.abc import Sequence
from pathlib import Path

from kraken.common import flatten
from kraken.core import Project, Property, TaskStatus

from .base_task import EnvironmentAwareDispatchTask


class CoverageFormat(enum.Enum):
    XML = ("xml", ".xml")
    HTML = ("html", "_html")
    JSON = ("json", ".json")
    LCOV = ("lcov", ".info")
    ANNOTATE = ("annotate", "_annotate")

    def get_format(self) -> str:
        return self.value[0]

    def get_suffix(self) -> str:
        return self.value[1]


class PytestTask(EnvironmentAwareDispatchTask):
    description = "Run unit tests using Pytest."
    python_dependencies = ["pytest"]

    tests_dir: Property[Path]
    ignore_dirs: Property[Sequence[Path]] = Property.default_factory(list)
    allow_no_tests: Property[bool] = Property.default(False)
    doctest_modules: Property[bool] = Property.default(True)
    marker: Property[str]
    coverage: Property[CoverageFormat]

    # EnvironmentAwareDispatchTask

    def is_skippable(self) -> bool:
        return self.allow_no_tests.get() and self.tests_dir.is_empty() and not self.settings.get_tests_directory()

    def get_execute_command(self) -> list[str] | TaskStatus:
        tests_dir = self.tests_dir.get_or(None)
        tests_dir = tests_dir or self.settings.get_tests_directory()
        if not tests_dir:
            print("error: no test directory configured and none could be detected")
            return TaskStatus.failed("no test directory configured and none could be detected")
        command = [
            "pytest",
            "-vv",
            str(self.project.directory / self.settings.source_directory),
            str(self.project.directory / tests_dir),
        ]
        command += flatten(["--ignore", str(self.project.directory / path)] for path in self.ignore_dirs.get())
        command += ["--log-cli-level", "INFO"]
        if self.coverage.is_filled():
            coverage_file = f"coverage{self.coverage.get().get_suffix()}"
            command += [
                "--cov-report",
                f"{self.coverage.get().get_format()}:{str(self.project.build_directory / coverage_file)}",
                "--cov-report",
                "term",
                f"--cov={str(self.project.directory / self.settings.source_directory)}",
            ]
        if self.marker.is_filled():
            command += ["-m", self.marker.get()]
        if self.doctest_modules.get():
            command += ["--doctest-modules"]
        command += shlex.split(os.getenv("PYTEST_FLAGS", ""))
        return command

    def handle_exit_code(self, code: int) -> TaskStatus:
        if code == 5 and self.allow_no_tests.get():
            # Pytest returns exit code 5 if no tests were run.
            return TaskStatus.succeeded()
        return TaskStatus.from_exit_code(None, code)


def pytest(
    *,
    name: str = "pytest",
    group: str = "test",
    project: Project | None = None,
    tests_dir: Path | None = None,
    ignore_dirs: Sequence[Path] = (),
    allow_no_tests: bool = False,
    doctest_modules: bool = True,
    marker: str | None = None,
    coverage: CoverageFormat | None = None,
) -> PytestTask:
    project = project or Project.current()
    task = project.task(name, PytestTask, group=group)
    task.tests_dir = tests_dir
    task.ignore_dirs = ignore_dirs
    task.allow_no_tests = allow_no_tests
    task.doctest_modules = doctest_modules
    task.marker = marker
    task.coverage = coverage
    return task
